""" Task Queue for DAX in REDCap."""

import os
import logging
import json

import pandas as pd

from dax import XnatUtils


logger = logging.getLogger('manager.rcq.taskqueue')


class TaskQueue(object):

    def __init__(self, projects_redcap):
        self._rc = projects_redcap

    def sync(self, xnat, projects):
        task_updates = []
        def_field = self._rc.def_field

        logger.info('loading taskqueue records from REDCap')
        rec = self._rc.export_records(
            records=projects,
            forms=['taskqueue'],
            fields=[def_field])

        # Filter to only uploading jobs
        rec = [x for x in rec if x['redcap_repeat_instrument'] == 'taskqueue']
        rec = [x for x in rec if x['task_status'] in ['LOST', 'UPLOADING']]

        if len(rec) == 0:
            logger.info('no active tasks to update')
            return

        # Get projects with active tasks
        projects = list(set([x[def_field] for x in rec]))

        # Load assesssor data from XNAT as our current status
        logger.info(f'loading XNAT data:{projects}')
        assr_data = XnatUtils.load_assr_data(xnat, projects)
        sgp_data = XnatUtils.load_sgp_data(xnat, ','.join(projects))

        # Get updates
        logger.debug('updating each task')
        for i, t in enumerate(rec):
            assr = t['task_assessor']
            task_status = t['task_status']

            # Find the matching assessor for this task in XNAT data
            match_found = False
            for k, a in pd.concat([assr_data, sgp_data]).iterrows():
                if a.ASSR != assr:
                    continue

                match_found = True
                xnat_status = a.PROCSTATUS

                # We have a match, now check for updates
                if xnat_status == 'COMPLETE':
                    logger.debug(f'{i}:{assr}:COMPLETE')

                    # Apply complete from xnat to redcap
                    task_updates.append({
                        def_field: t[def_field],
                        'redcap_repeat_instrument': 'taskqueue',
                        'redcap_repeat_instance': t['redcap_repeat_instance'],
                        'task_status': 'COMPLETE',
                        'taskqueue_complete': '2'
                    })
                elif xnat_status == 'JOB_FAILED':
                    logger.debug(f'{i}:{assr}:JOB_FAILED')

                    # Apply from xnat to redcap
                    task_updates.append({
                        def_field: t[def_field],
                        'redcap_repeat_instrument': 'taskqueue',
                        'redcap_repeat_instance': t['redcap_repeat_instance'],
                        'task_status': 'JOB_FAILED',
                        'taskqueue_complete': '0'
                    })
                elif task_status == 'LOST':
                    logger.info(f'found lost job, setting to QUEUED:{assr}')

                    task_updates.append({
                        def_field: t[def_field],
                        'redcap_repeat_instrument': 'taskqueue',
                        'redcap_repeat_instance': t['redcap_repeat_instance'],
                        'task_status': 'QUEUED',
                        'task_jobid': '',
                        'task_jobnode': '',
                    })
                else:
                    logger.debug(f'{i}:{assr}:{task_status}')

                # we matched so we are done with this task
                break

            if not match_found:
                logger.debug(f'no match, set to DELETED:{assr}')
                task_updates.append({
                    def_field: t[def_field],
                    'redcap_repeat_instrument': 'taskqueue',
                    'redcap_repeat_instance': t['redcap_repeat_instance'],
                    'task_status': 'DELETED',
                })

        # Apply updates to REDCap
        logger.debug(f'updating {len(task_updates)} tasks')

        if task_updates:
            logger.debug(f'updating task info on redcap:{assr}:{task_updates}')
            self.apply_updates(task_updates)
        else:
            logger.debug(f'nothing to update')

    def apply_updates(self, updates):
        try:
            self._rc.import_records(updates)
        except Exception as err:
            logger.error(err)

    def _assessor_task_id(self, project, assessor):
        task_id = None

        rec = self._rc.export_records(
            forms=['taskqueue'],
            records=[project],
            fields=[self._rc.def_field, 'task_assessor'])

        rec = [x for x in rec if x['redcap_repeat_instrument'] == 'taskqueue']
        rec = [x for x in rec if x['task_assessor'] == assessor]

        if len(rec) > 1:
            logger.warn(f'duplicate tasks for assessor, not good:{assessor}')
            task_id = rec[0]['redcap_repeat_instance']
        elif len(rec) == 1:
            task_id = rec[0]['redcap_repeat_instance']

        return task_id

    def current_tasks(self):
        return

    def _add_task(
        self,
        project,
        assr,
        inputlist,
        var2val,
        walltime,
        memreq,
        yamlfile,
        userinputs,
        custom=False
    ):
        """Add a new task record ."""

        # Convert to string for storing
        var2val = json.dumps(var2val)
        inputlist = json.dumps(inputlist)
        userinputs = json.dumps(userinputs)
        def_field = self._rc.def_field

        # Try to match existing record
        task_id = self._assessor_task_id(project, assr)

        if custom:
            task_yamlfile = 'CUSTOM'
        else:
            task_yamlfile = os.path.basename(yamlfile)

        if task_id:
            # Update existing record
            try:
                record = {
                    def_field: project,
                    'redcap_repeat_instrument': 'taskqueue',
                    'redcap_repeat_instance': task_id,
                    'task_status': 'QUEUED',
                    'task_inputlist': inputlist,
                    'task_var2val': var2val,
                    'task_walltime': walltime,
                    'task_memreq': memreq,
                    'task_yamlfile': task_yamlfile,
                    'task_userinputs': userinputs,
                    'task_timeused': '',
                    'task_memused': '',
                }
                response = self._rc.import_records([record])
                assert 'count' in response
                logger.debug('task record created')
            except AssertionError as err:
                logger.error(f'upload failed:{err}')
                return
        else:
            # Create a new record
            try:
                record = {
                    def_field: project,
                    'redcap_repeat_instrument': 'taskqueue',
                    'redcap_repeat_instance': 'new',
                    'task_assessor': assr,
                    'task_status': 'QUEUED',
                    'task_inputlist': inputlist,
                    'task_var2val': var2val,
                    'task_walltime': walltime,
                    'task_memreq': memreq,
                    'task_yamlfile': task_yamlfile,
                    'task_userinputs': userinputs,
                }
                response = self._rc.import_records([record])
                assert 'count' in response
                logger.debug('task record created')

            except AssertionError as err:
                logger.error(f'upload failed:{err}')
                return

        # If the file is not in yaml dir, we need to upload it to the task
        if task_yamlfile == 'CUSTOM':
            logger.debug(f'yaml not in shared library, uploading to task')
            if not task_id:
                # Try to match existing record
                task_id = self._assessor_task_id(project, assr)

            logger.debug(f'uploading file:{yamlfile}')
            self._upload_task_processor_file(
                project,
                task_id,
                yamlfile)

    def _upload_task_processor_file(self, record_id, repeat_id, filename):
        with open(filename, 'rb') as f:
            return self._rc.import_file(
                record=record_id,
                field='task_yamlupload',
                file_name=os.path.basename(filename),
                event=None,
                repeat_instance=repeat_id,
                file_object=f)



import yamale
import os

from .errors import DaxError


def validate(filename):
    # TODO: match file to schema by loading yaml, check procyamlversion, etc

    # Load the schema
    try:
        schema_file = os.path.realpath(os.path.join(
            os.path.dirname(__file__), 'schema', 'processor.yaml'))

        schema = yamale.make_schema(schema_file)
    except Exception as err:
        raise DaxError('failed to read schema:{}:{}'.format(schema_file, err))

    # Load the file to be validated
    try:
        data = yamale.make_data(filename)
    except Exception as err:
        raise DaxError('failed to read file:{}:{}'.format(filename, err))

    # Validate data against the schema
    try:
        yamale.validate(schema, data)
    except ValueError as err:
        raise DaxError('validate failed:{}:{}'.format(schema_file, err))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" cluster.py

Cluster functionality
"""

import os
import time
import logging
import subprocess as sb
from datetime import datetime

from .dax_settings import DAX_Settings
from .errors import ClusterError


__copyright__ = 'Copyright 2013 Vanderbilt University. All Rights Reserved'
DAX_SETTINGS = DAX_Settings()
MAX_TRACE_DAYS = 30
LOGGER = logging.getLogger('dax')


def c_output(output):
    """
    Check if the output value is an integer

    :param output: variable to check
    :return: True if output is not an integer, False otherwise.
    """
    try:
        int(output)
        error = False
    except ValueError as err:
        error = True
        LOGGER.error(err)
    return error


def cj_subcmd(cmd):
    output = sb.check_output(cmd, shell=True)
    error = c_output(output)
    while error:
        LOGGER.info('    try again to access number of jobs in 2 seconds.')
        time.sleep(2)
        output = sb.check_output(cmd, shell=True)
        error = c_output(output)
    if int(output) < 0:
        return 0
    else:
        return int(output)


def count_jobs(resdir,force_no_qsub=False):
    """
    Count the number of jobs in the queue on the cluster

    :return: number of jobs in the queue. tuple of
             (launched, pending, pendinguploads)
    """
    if force_no_qsub:
        LOGGER.info(' Running locally. No queue with jobs.')
        return (0, 0, 0)
    elif command_found(cmd=DAX_SETTINGS.get_cmd_submit()):
        launched = cj_subcmd(DAX_SETTINGS.get_cmd_count_jobs_launched())
        pending = cj_subcmd(DAX_SETTINGS.get_cmd_count_jobs_pending())
        cmd = DAX_SETTINGS.get_cmd_count_pendinguploads()
        cmd = cmd.safe_substitute({'resdir': resdir})
        pendinguploads = cj_subcmd(cmd)
        return (launched, pending, pendinguploads)
    else:
        LOGGER.error('ERROR: failed to find cluster commands')
        raise ClusterLaunchException


def job_status(jobid):
    """
    Get the status for a job on the cluster

    :param jobid: job id to check
    :return: job status

    """
    cmd = DAX_SETTINGS.get_cmd_get_job_status()\
                      .safe_substitute({'jobid': jobid})

    LOGGER.debug(str(cmd).strip())

    try:
        output = sb.check_output(cmd, stderr=sb.STDOUT, shell=True)
        LOGGER.debug('output='+str(output))
        output = output.decode().strip()
        if output == DAX_SETTINGS.get_running_status():
            return 'R'
        elif output == DAX_SETTINGS.get_queue_status():
            return 'Q'
        elif output == DAX_SETTINGS.get_complete_status() or len(output) == 0:
            return 'C'
        else:
            return None
    except sb.CalledProcessError as e:
        LOGGER.debug(str(e))
        return None


def is_traceable_date(jobdate):
    """
    Check if the job is traceable on the cluster

    :param jobdate: launching date of the job
    :return: True if traceable, False otherwise.
    """
    try:
        trace_date = datetime.strptime(jobdate, "%Y-%m-%d")
        diff_days = (datetime.today() - trace_date).days
        return diff_days <= MAX_TRACE_DAYS
    except ValueError:
        return False


def tracejob_info(jobid, jobdate):
    """
    Trace the job information from the cluster

    :param jobid: job id to check
    :param jobdate: launching date of the job
    :return: dictionary object with 'mem_used', 'walltime_used', 'jobnode'
    """
    time_s = datetime.strptime(jobdate, "%Y-%m-%d")
    diff_days = (datetime.today() - time_s).days + 1
    jobinfo = dict()
    jobinfo['mem_used'] = get_job_mem_used(jobid, diff_days)
    jobinfo['walltime_used'] = get_job_walltime_used(jobid, diff_days)
    jobinfo['jobnode'] = get_job_node(jobid, diff_days)

    return jobinfo


def get_job_mem_used(jobid, diff_days):
    """
    Get the memory used for the task from cluster

    :param jobid: job id to check
    :param diff_days: difference of days between starting date and now
    :return: string with the memory usage, empty string if error
    """
    mem = ''
    # Check for blank jobid
    if not jobid:
        return mem

    cmd = DAX_SETTINGS.get_cmd_get_job_memory()\
                      .safe_substitute({'numberofdays': diff_days,
                                        'jobid': jobid})
    try:
        output = sb.check_output(cmd, stderr=sb.STDOUT, shell=True)
        if output.startswith(b'sacct: error'):
            raise ClusterError(output)
        if output:
            mem = output.strip()
            mem = mem.decode()

    except (sb.CalledProcessError, ClusterError):
        pass

    return mem


def get_job_walltime_used(jobid, diff_days):
    """
    Get the walltime used for the task from cluster

    :param jobid: job id to check
    :param diff_days: difference of days between starting date and now
    :return: string with the walltime used, empty string if error
    """
    walltime = ''
    # Check for blank jobid
    if not jobid:
        return walltime

    cmd = DAX_SETTINGS.get_cmd_get_job_walltime()\
                      .safe_substitute({'numberofdays': diff_days,
                                        'jobid': jobid})

    try:
        output = sb.check_output(cmd, stderr=sb.STDOUT, shell=True)
        if output:
            walltime = output.strip()
            walltime = walltime.decode()

    except sb.CalledProcessError:
        pass

    if not walltime and diff_days > 3:
        walltime = 'NotFound'

    return walltime


def get_job_node(jobid, diff_days):
    """
    Get the node where the job was running on the cluster

    :param jobid: job id to check
    :param diff_days: difference of days between starting date and now
    :return: string with the node, empty string if error
    """
    jobnode = ''
    # Check for blank jobid
    if not jobid:
        return jobnode

    if jobid == 'no_qsub':
        cmd = 'uname -a'
        output = sb.check_output(cmd, stderr=sb.STDOUT, shell=True)
        if output and len(output.strip().split(' ')) > 1:
            jobnode = output.strip().split(' ')[1]
        return jobnode

    cmd = DAX_SETTINGS.get_cmd_get_job_node()\
                      .safe_substitute({'numberofdays': diff_days,
                                        'jobid': jobid})

    try:
        output = sb.check_output(cmd, stderr=sb.STDOUT, shell=True)
        if output:
            jobnode = output.strip()
            jobnode = jobnode.decode()
    except sb.CalledProcessError:
        pass

    return jobnode


def get_specific_str(big_str, prefix, suffix):
    """
    Extract a specific length out of a string

    :param big_str: string to reduce
    :param prefix: prefix to remove
    :param suffix: suffix to remove
    :return: string reduced, return empty string if prefix/suffix not present
    """
    specific_str = big_str
    if prefix and len(specific_str.split(prefix)) > 1:
        specific_str = specific_str.split(prefix)[1]
    if suffix and len(specific_str.split(suffix)) > 1:
        specific_str = specific_str.split(suffix)[0]
    if specific_str != big_str:
        return specific_str
    else:
        return ''


def command_found(cmd='qsub'):
    """ Return True if the command was found."""
    if True in [os.path.isfile(os.path.join(path, cmd)) and
                os.access(os.path.join(path, cmd), os.X_OK)
                for path in os.environ["PATH"].split(os.pathsep)]:
        return True
    return False


class PBS(object):   # The script file generator class
    """ PBS class to generate/submit the cluster file to run a task """
    def __init__(self, filename, outfile, cmds, walltime_str, mem_mb=2048,
                 ppn=1, env=None, email=None,
                 email_options=DAX_SETTINGS.get_email_opts(), rungroup=None,
                 xnat_host=None, job_template=None):
        """
        Entry point for the PBS class

        :param filename: filename for the script
        :param outfile: filepath for the outlogs
        :param cmds: commands to run in the script
        :param walltime_str: walltime to set for the script
        :param mem_mb: memory in mb to set for the script
        :param ppn: number of processor to set for the script
        :param env: Environment file to source  for the script
        :param email: email address to set for the script
        :param email_options: email options to set for the script
        :param rungroup: group to run job under on the cluster
        :param xnat_host: set the XNAT_HOST for the job (export)
        :return: None
        """
        self.filename = filename
        self.outfile = outfile
        self.cmds = cmds
        self.walltime_str = walltime_str
        self.mem_mb = mem_mb
        self.email = email
        self.email_options = email_options
        self.rungroup = rungroup
        self.ppn = ppn
        self.job_template = job_template
        if env:
            self.env = env
        else:
            self.env = os.path.join(os.environ['HOME'], '.bashrc')
        if xnat_host:
            self.xnat_host = xnat_host
        else:
            self.xnat_host = os.environ['XNAT_HOST']

    def write(self):
        """
        Write the file

        :return: None
        """
        # pbs_dir
        job_dir = os.path.dirname(self.filename)
        if not os.path.exists(job_dir):
            os.makedirs(job_dir)

        job_data = {'job_email': self.email,
                    'job_email_options': self.email_options,
                    'job_rungroup': self.rungroup,
                    'job_ppn': str(self.ppn),
                    'job_env': str(self.env),
                    'job_walltime': str(self.walltime_str),
                    'job_memory': str(self.mem_mb),
                    'job_output_file': self.outfile,
                    'job_output_file_options': 'oe',
                    'job_cmds': '\n'.join(self.cmds),
                    'xnat_host': self.xnat_host}

        # Load job template
        try:
            _tmp = DAX_SETTINGS.get_job_template(self.job_template)
            _str = _tmp.safe_substitute(job_data)
        except Exception as err:
            raise ClusterError(f'failed to read job template:{err}')

        # Write the file
        with open(self.filename, 'w') as f_obj:
            f_obj.write(_str)

    def submit(self, outlog=None, force_no_qsub=False):
        """
        Submit the file to the cluster

        :return: None
        """
        return submit_job(self.filename, outlog=outlog,
                          force_no_qsub=force_no_qsub)


def submit_job(filename, outlog=None, force_no_qsub=False):
    """
    Submit the file to the cluster
    :return: jobid and error if the job failed when running locally
    """
    failed = False
    submit_cmd = DAX_SETTINGS.get_cmd_submit()
    if command_found(cmd=submit_cmd) and not force_no_qsub:
        try:
            cmd = '%s %s' % (submit_cmd, filename)
            proc = sb.Popen(cmd.split(), stdout=sb.PIPE, stderr=sb.PIPE)
            output, error = proc.communicate()
            if output:
                LOGGER.info(output.decode().strip())
            if error:
                LOGGER.error(error.decode())
            jobid = get_specific_str(
                output.decode(), DAX_SETTINGS.get_prefix_jobid(),
                DAX_SETTINGS.get_suffix_jobid())
        except sb.CalledProcessError as err:
            LOGGER.error(err)
            jobid = '0'

    else:
        cmd = 'sh %s' % (filename)
        proc = sb.Popen(cmd.split(), stdout=sb.PIPE, stderr=sb.PIPE)
        output, error = proc.communicate()
        if outlog:
            with open(outlog, 'w') as log_obj:
                for line in output:
                    log_obj.write(line)
                for line in error:
                    log_obj.write(line)

        if error:
            # Set the status to JOB_FAILED
            failed = True
        jobid = 'no_qsub'

    return jobid.strip(), failed



#!/usr/bin/env python

import json
import sys
import pandas
import pyxnat
from dax import XnatUtils
from dax import utilities

if len(sys.argv)!=4 :
    print('Usage:')
    print('python Xnatreport_assessor.py <project> <proctype> <output_csvfile>')
    sys.exit()


project = sys.argv[1]
proctype = sys.argv[2]
outfile = sys.argv[3]

xnat = XnatUtils.get_interface()

Assrs = xnat.list_project_assessors(project)

R = list()
for assr in Assrs :

    if assr.get('proctype') != proctype : continue

    #print(assr['assessor_label'])

    # Get desired fields
    thisR = {}
    for key in ('project_label','subject_label','session_label','proctype',
                'assessor_id','procstatus','qcstatus','assessor_label') :
        thisR[key] = assr[key]

    # Clean up the inputs field, split on / and keep the last bit
    inps = utilities.decode_url_json_string(assr['inputs'])
    #inps = json.loads(assr['inputs'].replace('&quot;','"'))
    for key in inps.keys() :
        thisR[key] = inps[key].split('/')[-1]
    
    # We need to explicitly copy here to avoid overwriting R
    R.append(thisR.copy())


D = pandas.DataFrame(R)

# Reorder columns
colorder = ('project_label','subject_label','session_label','proctype',
            'assessor_id','procstatus','qcstatus')
oldcols = D.columns.tolist()
newcols = list()
for col in colorder :
    newcols.append(col)
    oldcols.remove(col)
newcols.extend(oldcols)

D.to_csv(outfile,index=False,columns=newcols)





from unittest import TestCase

import itertools


from dax import XnatUtils


class InterfaceTempUnitTests(TestCase):

    def unit_test_connection_strings(self):
        #positive tests
        self.assertEqual(
            XnatUtils.InterfaceTemp.P_XPATH.format(project='proj1'),
            '/project/proj1'
        )

        self.assertEqual(
            XnatUtils.InterfaceTemp.S_XPATH.format(project='proj1',
                                                   subject='subj1'),
            '/project/proj1/subject/subj1'
        )

        self.assertEqual(
            XnatUtils.InterfaceTemp.E_XPATH.format(project='proj1',
                                                   subject='subj1',
                                                   session='sess1'),
            '/project/proj1/subject/subj1/experiment/sess1'
        )

        self.assertEqual(
            XnatUtils.InterfaceTemp.C_XPATH.format(project='proj1',
                                                   subject='subj1',
                                                   session='sess1',
                                                   scan='scan1'),
            '/project/proj1/subject/subj1/experiment/sess1/scan/scan1'
        )

        self.assertEqual(
            XnatUtils.InterfaceTemp.CR_XPATH.format(project='proj1',
                                                    subject='subj1',
                                                    session='sess1',
                                                    scan='scan1',
                                                    resource='res1'),
            '/project/proj1/subject/subj1/experiment/sess1'
            '/scan/scan1/resource/res1'
        )

        self.assertEqual(
            XnatUtils.InterfaceTemp.A_XPATH.format(project='proj1',
                                                   subject='subj1',
                                                   session='sess1',
                                                   assessor='assr1'),
            '/project/proj1/subject/subj1/experiment/sess1/assessor/assr1'
        )

        self.assertEqual(
            XnatUtils.InterfaceTemp.AR_XPATH.format(project='proj1',
                                                    subject='subj1',
                                                    session='sess1',
                                                    assessor='assr1',
                                                    resource='res1'),
            '/project/proj1/subject/subj1/experiment/sess1'
            '/assessor/assr1/out/resource/res1'

        )

    def test_object_type_from_path(self):
        tests = [
            ('project/p1', 'project'),
            ('project/p1/subject/s1', 'subject'),
            ('project/p1/subject/s1/experiment/e1', 'experiment'),
            ('project/p1/subject/s1/experiment/e1/scan/sc1', 'scan'),
            ('project/p1/subject/s1/experiment/e1/assessor/as1', 'assessor'),
            ('project/p1/subject/s1/experiment/e1/scan/sc1/resource/r1',
             'resource'),
            ('project/p1/subject/s1/experiment/e1/assessor/as1/in/resource/r1',
             'resource'),
            ('project/p1/subject/s1/experiment/e1/assessor/as1/out/resource/r1',
             'resource'),
        ]

        prefix = ['', 'data/', 'xnat:/', '/', '/data/']
        postfix = ['', '/']

        for prepost in itertools.product(prefix, postfix):
            for t in tests:
                instr = prepost[0] + t[0] + prepost[1]
                print(('testing ', instr))
                self.assertEqual(
                    t[1],
                    XnatUtils.InterfaceTemp.object_type_from_path(instr),
                    'unexpected object type')




import json

from dax import XnatUtils


class SessionTools:

    @staticmethod
    def get_connection(host):
        return XnatUtils.get_interface(host=host)

    @staticmethod
    def prep_project(intf, proj_id, subj_id, sess_id, scans, assessors):
        sess = intf.select_experiment(proj_id, subj_id, sess_id)
        for scan in scans:
            SessionTools.add_scan(sess, scan['name'], scan)
        for assessor in assessors:
            SessionTools.add_assessor(assessor, assessor['name'], assessor)

    @staticmethod
    def add_session(proj, name, parameters):
        sess = proj.scan(name)
        if sess.exists():
            sess.delete()
            sess = proj.session(name)
        sess.create(session=parameters['xsitype'])
        return sess

    @staticmethod
    def add_scan(sess, name, parameters):
        scn = sess.scan(name)
        if scn.exists():
            scn.delete()
            scn = sess.scan(name)
        kwargs = dict()
        print(parameters)
        print((parameters['xsitype']))
        kwargs[parameters['xsitype'] + '/type'] = parameters['type']
        kwargs[parameters['xsitype'] + '/quality'] = parameters['quality']
        scn.create(scans=parameters['xsitype'], **kwargs)
        for output in parameters['files']:
            for f in output[1]:
                scn.resource(output[0]).file(f).insert('./file.txt')
                # print output[0], f
        return scn

    @staticmethod
    def add_assessor(sess, name, parameters, inputs_policy="empty_if_not_set"):
        asr = sess.assessor(name)
        if asr.exists():
            asr.delete()
            asr = sess.assessor(name)
        kwargs = dict()
        kwargs[parameters['xsitype'] + '/proctype'] = parameters['proctype']
        kwargs[parameters['xsitype'] + '/procversion'] = '1.0.0'
        kwargs[parameters['xsitype'] + '/validation/status'] = "Needs QA"
        if inputs_policy == "empty_if_not_set":
            kwargs[parameters['xsitype'] + '/inputs'] =\
                json.dumps(parameters.get('inputs', {}))
        elif inputs_policy == "not_inputs":
            pass

        asr.create(assessors=parameters['xsitype'], **kwargs)
        for output in parameters['files']:
            for f in output[1]:
                asr.resource(output[0]).file(f).insert('./file.txt')
        return asr



# TODO: daxlauncher.upload()

import logging

from dax import bin, XnatUtils


# Create the launcher
settingspath = '/Users/boydb1/.dax/settings/settings-subjgenproc.yaml'
daxlauncher = bin.read_yaml_settings(settingspath, logging.getLogger())

# Get xnat connection
xnat = XnatUtils.get_interface(host=daxlauncher.xnat_host)

# Build
print('building')
project = 'DepMIND2'
daxlauncher.build_project_subjgenproc(xnat, project)

print('All Done!')



""" rcq implements a job queue for DAX in REDCap. Launch and Finish."""

import os
import logging
import subprocess
from datetime import datetime

import yaml
import requests

from ..cluster import PBS, count_jobs
from ..lockfiles import lock_flagfile, unlock_flagfile
from .projectinfo import load_project_info
from ..utilities import get_this_instance


# Handle analysis jobs that only need to be launched/updated, but not uploaded.
# The job will upload it's own output. On finish, we update job information on
# redcap and delete log and batch files from disk. No job information is stored
# on disk, no diskq, rcq.


logger = logging.getLogger('manager.rcq.analysislauncher')


DONE_STATUSES = ['COMPLETE', 'JOB_FAILED', 'READY', 'DEVEL']


CMDS_TEMPLATE = '''

ANALYSISID={analysis}
PROJECT={project}
REPO={repo}
VERSION={version}
INLIST={inputs}
XNATHOST={host}
XNATUSER={user}
{maincmd}

'''


SINGULARITY_BASEOPTS = (
    '-e '
    '--home $JOBDIR:$HOME '
    '--bind $REPODIR:/REPO '
    '--bind $INDIR:/INPUTS '
    '--bind $OUTDIR:/OUTPUTS '
    '--bind $JOBDIR:/tmp '
    '--bind $JOBDIR:/dev/shm '
)


class AnalysisLauncher(object):
    def __init__(self, xnat, projects_redcap, instance_settings):
        self._xnat = xnat
        self._projects_redcap = projects_redcap
        self._instance_settings = instance_settings
        self._resdir = self._instance_settings['main_resdir']
        self._rcqdir = f'{self._resdir}/RCQ'
        self._job_template = None

    def _get_job_template(self):
        if self._job_template is None:
            # Determine job template
            self._job_template = self._instance_settings.get(
                'main_projectjobtemplate', None)

        if self._job_template is None:
            #_job_template = instance_settings.get('jobtemplate', None)
            #job_template = f'{_job_template}/project_job_template.txt'
            self._job_template = os.path.expanduser(
                '~/project_job_template.txt')

        return self._job_template

    def load_processor_redcap(self, project, repeat_id):
        """Export file from REDCap."""

        # Get the file contents from REDCap
        try:
            (cont, hdr) = self._projects_redcap.export_file(
                record=project,
                field='analysis_processor',
                repeat_instance=repeat_id)

            if cont == '':
                raise Exception('error exporting file from REDCap')
        except Exception as err:
            logger.error(f'downloading file:{err}')
            return None

        return cont

    def load_processor_github(self, user, repo, version, subdir):
        """Export file from github."""

        base = 'https://raw.githubusercontent.com'
        filename = 'processor.yaml'

        if subdir:
            url = f'{base}/{user}/{repo}/{version}/processors/{subdir}/{filename}'
        else:
            url = f'{base}/{user}/{repo}/{version}/{filename}'

        logger.info(f'{url=}')

        # Get the file contents
        try:
            r = requests.get(url, allow_redirects=True)
            if r.content == '':
                raise Exception('error exporting file from github')
        except Exception as err:
            logger.error(f'downloading file:{err}')
            return None

        return yaml.safe_load(r.text)

    def update(self, projects, launch_enabled=True):
        """Update all analyses in projects_redcap."""
        launch_list = []
        updates = []
        def_field = self._projects_redcap.def_field
        projects_redcap = self._projects_redcap
        instance_settings = self._instance_settings

        # Lock rcq for updating
        lock_file = f'{self._resdir}/FlagFiles/rcq.pid'
        success = lock_flagfile(lock_file)
        if not success:
            # Failed to get lock
            logger.warn('failed to get lock:{}'.format(lock_file))
            return

        try:
            # Get the current analysis table
            logger.debug('loading current analysis records')
            try:
                rec = projects_redcap.export_records(
                    records=projects,
                    forms=['analyses'],
                    fields=[def_field])
            except Exception as err:
                logger.error(f'failed to load analyses:{err}')
                return

            rec = [x for x in rec if x['redcap_repeat_instrument'] == 'analyses']

            logger.debug(f'Found {len(rec)} analysis records')

            # Filter to only open jobs
            rec = [x for x in rec if x['analysis_status'] not in DONE_STATUSES]

            logger.debug(f'{len(rec)} analysis with open jobs')

            # Update each record
            for i, cur in enumerate(rec):
                project = cur[def_field]
                cur['project'] = project
                instance = cur['redcap_repeat_instance']
                status = cur['analysis_status']

                logger.debug(f'{i}:{project}:{instance}:{status}')

                if status in ['NEED_INPUTS', 'UPLOADING', 'READY', 'DEVEL']:
                    logger.debug(f'ignoring status={status}')
                    pass
                elif status == 'RUNNING':
                    # check on running job
                    logger.debug('checking on running job')
                    cur_updates = get_updates(cur)
                    if cur_updates:
                        cur_updates.update({
                            def_field: cur[def_field],
                            'redcap_repeat_instrument': 'analyses',
                            'redcap_repeat_instance': instance,
                        })

                        # Add to redcap updates
                        updates.append(cur_updates)
                elif status in ['COMPLETED', 'FAILED', 'CANCELLED']:
                    # finish completed job after being uploaded to xnat
                    logger.debug(f'handling complete, status={status}')

                    # Get the output label on xnat
                    label = cur.get('analysis_output', None)
                    if not label:
                        logger.error('label not found, cannot finish')
                        continue

                    logger.debug(f'{label=}')

                    # Upload slurm file to xnat
                    slurm_file = f'{self._rcqdir}/{label}.slurm'
                    if os.path.isfile(slurm_file):
                        logger.debug(f'upload slurm file:{slurm_file}')
                        self._upload_file(slurm_file, project, label)
                        os.remove(slurm_file)

                    # Upload log file to xnat
                    log_file = f'{self._rcqdir}/{label}.txt'
                    if os.path.isfile(log_file):
                        logger.debug(f'upload log file:{log_file}')
                        self._upload_file(log_file, project, label)
                        os.remove(log_file)

                    # Finalize the status from xnat
                    if status == 'COMPLETED':
                        status = 'READY'
                    elif status == 'FAILED':
                        status = 'JOB_FAILED'
                    elif status == 'CANCELLED':
                        status = 'JOB_CANCELLED'

                    # Add to redcap updates
                    updates.append({
                        def_field: cur[def_field],
                        'redcap_repeat_instrument': 'analyses',
                        'redcap_repeat_instance': instance,
                        'analysis_status': status,
                    })
                elif status == 'QUEUED' or status == 'Q':
                    logger.debug('adding queued job to launch list')
                    launch_list.append(cur)
                else:
                    logger.info(f'unknown status:{status}')
                    continue

            if updates:
                # upload changes
                logger.debug(f'updating redcap:{updates}')
                try:
                    projects_redcap.import_records(updates)
                except Exception as err:
                    err = 'connection to REDCap interrupted'
                    logger.error(err)

            if launch_enabled:

                # Make the rcq dir for slurm scripts and output logs
                try:
                    os.makedirs(self._rcqdir)
                except FileExistsError:
                    pass

                q_limit = int(instance_settings['main_queuelimit'])
                p_limit = int(instance_settings['main_queuelimit_pending'])
                u_limit = int(instance_settings['main_limit_pendinguploads'])

                # Launch jobs
                updates = []
                for i, cur in enumerate(launch_list):
                    launched, pending, uploads = count_jobs(self._resdir)
                    logger.info(f'Cluster:{launched}/{q_limit} total, {pending}/{p_limit} pending, {uploads}/{u_limit} uploads')

                    if launched >= q_limit:
                        logger.info(f'queue limit reached:{q_limit}')
                        break

                    if pending >= p_limit:
                        logger.info(f'pending limit reached:{p_limit}')
                        break

                    if uploads >= u_limit:
                        logger.info(f'upload limit reached:{u_limit}')
                        break

                    try:
                        # Launch it!
                        logger.debug(f'launch:{i}')
                        jobid, label = self.launch_analysis(cur)

                        # Check for success
                        if jobid and label:
                            # Append to updates for redcap, clear existing info
                            updates.append({
                                def_field: cur[def_field],
                                'redcap_repeat_instrument': 'analyses',
                                'redcap_repeat_instance': cur['redcap_repeat_instance'],
                                'analysis_status': 'RUNNING',
                                'analysis_jobid': jobid,
                                'analysis_output': label,
                                'analysis_jobnode': 'NULL',
                                'analysis_timeused': 'NULL',
                                'analysis_memused': 'NULL',
                                'analysis_jobstarttime': 'NULL',
                                'analysis_jobendtime': 'NULL',
                            })
                    except Exception as err:
                        logger.error(err)
                        import traceback
                        traceback.print_exc()

                if updates:
                    # upload changes
                    logger.info(f'updating redcap:{updates}')
                    try:
                        projects_redcap.import_records(updates)
                    except Exception as err:
                        err = 'connection to REDCap interrupted'
                        logger.error(err)
        finally:
            # Delete the lock file
            logger.debug(f'deleting lock file:{lock_file}')
            unlock_flagfile(lock_file)

    def _upload_file(self, filename, project, analysis_label):
        res_uri = f'/projects/{project}/resources/{analysis_label}'
        logger.info(f'upload:{filename}:{res_uri}')
        self._xnat.select(res_uri).file(os.path.basename(filename)).put(
            filename,
            overwrite=True,
            params={"event_reason": "analysis upload"})

    def get_info(self, project):
        return load_project_info(self._xnat, project)

    def label_analysis(self, analysis):
        project = analysis['project']
        analysis_id = str(analysis['redcap_repeat_instance']).zfill(3)
        analysis_datetime = datetime.now().strftime("%Y-%m-%d-%H%M%S")
        analysis_label = f'{project}_A{analysis_id}_{analysis_datetime}'
        return analysis_label

    def get_inputlist(self, analysis):
        inputlist = []

        spec = analysis['processor']['inputs']['xnat']['subjects']

        projects = [analysis['project']]

        if analysis['analysis_otherprojects']:
            projects.extend(analysis['analysis_otherprojects'].split(','))

        for project in projects:
            info = self.get_info(project=project)

            subjects = analysis.get('subjects', None)
            if not subjects:
                # Get unique list of subjects from scans
                subjects = sorted(list(set([x['SUBJECT'] for x in info['scans']])))

            for subj in subjects:
                inputlist.extend(self.get_subject_inputs(spec, info, subj))

        # Prepend xnat host to each input path
        for i, d in enumerate(inputlist):
            _fpath = inputlist[i]['fpath']
            inputlist[i]['fpath'] = f'{self._xnat.host}/{_fpath}'

        return inputlist

    def get_session_inputs(self, spec, info, subject, session):
        inputs = []

        # Get the scans for this session
        scans = [x for x in info['scans'] if x['SESSION'] == session]

        for scan_spec in spec.get('scans', []):
            logger.debug(f'scan_spec={scan_spec}')

            # Get list of resources to download from this scan
            resources = scan_spec.get('resources', [])

            # Check for nifti tag
            if 'nifti' in scan_spec:
                # Add a NIFTI resource using value as fdest
                resources.append({
                    'resource': 'NIFTI',
                    'fdest': scan_spec['nifti']
                })

            scan_types = scan_spec['types'].split(',')

            logger.debug(f'scan_types={scan_types}')

            input_scans = [x for x in scans if x['SCANTYPE'] in scan_types]

            if len(input_scans) > 0 and scan_spec.get('keep_multis', '') != 'all':
                # Sort by id
                input_scans = sorted(input_scans, key=lambda x: x['SCANID'])
                num_scans = len(input_scans)

                # Apply keep_multis filter
                if scan_spec['keep_multis'] == 'first':
                    idx_multi = 1
                elif iv['keep_multis'] == 'last':
                    idx_multi = num_scans
                else:
                    try:
                        idx_multi = int(iv['keep_multis'])
                    except:
                        msg = 'keep_multis must be first, last, or index 1,2,3,...'
                        logger.error(msg)
                        raise Exception(msg)

                    if idx_multi > num_scans:
                        msg = f'{idx_multi} index exceeds found {num_scans}'
                        logger.error(msg)
                        raise Exception(msg)

                    # Get a list of only the requested scan
                    input_scans = [input_scans[idx_multi-1]]

            # Get the file inputs for each input scan
            for scan in input_scans:
                scanid = scan['SCANID']

                for res_spec in resources:
                    try:
                        res = res_spec['resource']
                    except (KeyError, ValueError) as err:
                        logger.error(f'reading resource:{err}')
                        continue

                    # Get the download destination subdir
                    ddest = f'{subject}/{session}/scans/{scanid}'
                    if res_spec.get('ddest', False):
                        ddest += '/' + res_spec.get('ddest')

                    if 'fdest' in res_spec:
                        _file = self._first_file(
                            info['name'],
                            subject,
                            session,
                            scan['SCANID'],
                            res
                        )
                        fpath = f'data/projects/{info["name"]}/subjects/{subject}/experiments/{session}/scans/{scanid}/resources/{res}/files/{_file}'
                        inputs.append(self._input(
                            fpath,
                            'FILE',
                            res_spec.get('fdest', _file),
                            ddest,
                        ))
                    elif 'fmatch' in res_spec:
                        for fmatch in res_spec['fmatch'].split(','):
                            fpath = f'data/projects/{info["name"]}/subjects/{subject}/experiments/{session}/scans/{scan["SCANID"]}/resources/{res}/files/{fmatch}'
                            inputs.append(self._input(
                                fpath,
                                'FILE',
                                res_spec.get('fdest', fmatch),
                                ddest
                            ))
                    else:
                        # Download whole resource
                        fpath = f'data/projects/{info["name"]}/subjects/{subject}/experiments/{session}/scans/{scan["SCANID"]}/resources/{res}/files'
                        inputs.append(self._input(
                            fpath,
                            'DIR',
                            None,
                            ddest
                        ))

        # Get the assessors for this session
        assessors = [x for x in info['assessors'] if x['SESSION'] == session]

        # Filter to only complete assessors
        logger.debug(f'found {len(assessors)} total assessors, filtering')
        assessors = [x for x in assessors if x['PROCSTATUS'] == 'COMPLETE']
        logger.debug(f'found {len(assessors)} complete assessors')

        for assr_spec in spec.get('assessors', []):
            logger.debug(f'assr_spec={assr_spec}')

            assr_types = assr_spec['types'].split(',')

            logger.debug(f'assr_types={assr_types}')

            for assr in [x for x in assessors if x['PROCTYPE'] in assr_types]:
                assrlabel = assr["ASSR"]

                for res_spec in assr_spec['resources']:

                    # Get the download destination subdir
                    ddest = f'{subject}/{session}/assessors/{assrlabel}'
                    if res_spec.get('ddest', False):
                        ddest += '/' + res_spec.get('ddest')

                    try:
                        res = res_spec['resource']
                    except (KeyError, ValueError) as err:
                        logger.error(f'reading resource:{err}')
                        continue

                    if 'fmatch' in res_spec:
                        for fmatch in res_spec['fmatch'].split(','):
                            fpath = f'data/projects/{info["name"]}/subjects/{subject}/experiments/{session}/assessors/{assrlabel}/out/resources/{res}/files/{fmatch}'
                            inputs.append(self._input(
                                fpath,
                                'FILE',
                                res_spec.get('fdest', fmatch),
                                ddest))
                    else:
                        # whole resource
                        fpath = f'data/projects/{info["name"]}/subjects/{subject}/experiments/{session}/assessors/{assrlabel}/out/resources/{res}/files'
                        inputs.append(self._input(
                            fpath,
                            'DIR',
                            res_spec.get('fdest', subject),
                            ddest))

        return inputs

    def get_subject_inputs(self, spec, info, subject):
        inputs = []

        # subject-level assessors, aka sgp
        if spec.get('assessors', None):
            logger.debug(f'get sgp:{subject}')
            sgp_spec = spec.get('assessors')
            sgp = [x for x in info['sgp'] if x['SUBJECT'] == subject]

            for assr in sgp:
                assrlabel = assr['ASSR']

                for assr_spec in sgp_spec:
                    logger.debug(f'assr_spec={assr_spec}')
                    assr_types = assr_spec['types'].split(',')
                    logger.debug(f'assr_types={assr_types}')

                    if assr['PROCTYPE'] not in assr_types:
                        logger.debug(
                            f'no match={assr["ASSR"]}:{assr["PROCTYPE"]}')
                        continue

                    for res_spec in assr_spec['resources']:

                        # Get the download destination subdir
                        ddest = f'{subject}/assessors/{assrlabel}'
                        if res_spec.get('ddest', False):
                            ddest += '/' + res_spec.get('ddest')

                        # Load the resource
                        try:
                            res = res_spec['resource']
                        except (KeyError, ValueError) as err:
                            logger.error(f'reading resource:{err}')
                            continue

                        if 'fmatch' in res_spec:
                            # Add each file
                            for fmatch in res_spec['fmatch'].split(','):
                                fpath = f'data/projects/{info["name"]}/subjects/{subject}/experiments/{assrlabel}/resources/{res}/files/{fmatch}'
                                inputs.append(self._input(
                                    fpath,
                                    'FILE',
                                    res_spec.get('fdest', fmatch),
                                    ddest
                                ))
                        else:
                            # We want the whole resource as one download
                            fpath = f'data/projects/{info["name"]}/subjects/{subject}/experiments/{assrlabel}/resources/{res}/files'
                            inputs.append(self._input(
                                fpath,
                                'DIR',
                                res_spec.get('fdest', subject),
                                ddest))

        # Download the subjects sessions
        for sess_spec in spec.get('sessions', []):

            if sess_spec.get('select', '') == 'first-mri':
                # only get the first mri
                subj_mris = [x for x in info['scans'] if (x['SUBJECT'] == subject and x['MODALITY'] == 'MR')]
                if len(subj_mris) < 1:
                    logger.debug('mri not found')
                    return

                # Sort by date and get first
                first = sorted(subj_mris, key=lambda x: x['DATE'])[0]

                # First session only
                sessions = [first['SESSION']]
            elif 'types' in sess_spec:
                # Find sessions matching types
                sess_types = sess_spec['types'].split(',')
                sessions = [x['SESSION'] for x in info['scans'] if x['SUBJECT'] == subject and x['SESSTYPE'] in sess_types]
                sessions = list(set(sessions))
            else:
                # Use any and all sessions for this subject
                sessions = [x['SESSION'] for x in info['scans'] if x['SUBJECT'] == subject]
                sessions = list(set(sessions))

            # Append inputs for this spec
            for sess in sessions:
                inputs += self.get_session_inputs(
                    sess_spec,
                    info,
                    subject,
                    sess
                )

        return inputs

    def _input(self, fpath, ftype, fdest=None, ddest=''):
        data = {
            'fpath': fpath,
            'ftype': ftype,
        }

        if fdest:
            data['fdest'] = fdest

        if ddest:
            data['ddest'] = ddest
        else:
            data['ddest'] = ''

        return data

    def _first_file(self, proj, subj, sess, scan, res):
        # Get name of the first file
        _files = self._xnat.select_scan_resource(
            proj, subj, sess, scan, res).files().get()
        return _files[0]

    def launch_analysis(self, analysis):
        """Launch as SLURM Job, write batch and submit. Return job id."""
        instance_settings = self._instance_settings

        if analysis['analysis_procrepo']:
            # Load the yaml file contents from github
            logger.info(f'loading:{analysis["analysis_procrepo"]}')
            p = analysis['analysis_procrepo'].replace(':', '/').split('/')
            if len(p) == 4:
                subdir = p[3]
            elif len(p) == 3:
                subdir = None
            else:
                logger.error(f'failed to parse:{analysis["analysis_procrepo"]}')
                return None

            user = p[0]
            repo = p[1]
            version = p[2]

            logger.info(f'loading:{user=}:{repo=}:{version=}:{subdir=}')

            analysis['processor'] = self.load_processor_github(
                user,
                repo,
                version,
                subdir=subdir
            )
            analysis['procrepo'] = f'https://github.com/{user}/{repo}/archive/refs/tags/{version}.tar.gz'
            analysis['procversion'] = version.replace('v', '')
        elif analysis['analysis_processor']:
            # Load the yaml file contents from REDCap
            logger.debug(f'loading:{analysis["analysis_processor"]}')
            analysis['processor'] = self.load_processor_redcap(
                analysis['project'],
                analysis['redcap_repeat_instance'])

        processor = analysis['processor']

        # Set subject list
        analysis['subjects'] = analysis['analysis_include'].splitlines()

        # Set the memory
        memreq = analysis.get('analysis_memreq', None)

        if not memreq:
            memreq = processor['requirements'].get('memory', None)

        if not memreq:
            memreq = '8G'

        # Set the walltime
        walltime = analysis.get('analysis_walltime', None)

        if walltime is None:
            walltime = processor['requirements'].get('walltime', None)

        if walltime is None:
            walltime = '0-12'

        # Create and set the label for our new analysis
        analysis['label'] = self.label_analysis(analysis)
        label = analysis['label']
        batch_file = f'{self._rcqdir}/{label}.slurm'
        log_file = f'{self._rcqdir}/{label}.txt'

        imagedir = instance_settings['main_singularityimagedir']
        xnat_host = instance_settings['main_xnathost']
        xnat_user = analysis.get('xnat_user', 'daxspider')

        # Build list of inputs
        inputlist = self.get_inputlist(analysis)

        # Get the job batch script template
        job_template = self._get_job_template()
        if not os.path.exists(job_template):
            logger.error('cannot find project job template')
            return

        # Set up containers
        for i, cur in enumerate(processor.get('containers', [])):
            # Get container path/filename
            cpath = cur['path']
            if not os.path.isabs(cpath) and imagedir:
                # Prepend singularity imagedir
                processor['containers'][i]['path'] = f'{imagedir}/{cpath}'

        cmds = self.get_cmds(xnat_host, xnat_user, inputlist, analysis)

        # Write the script
        logger.info(f'writing batch file:{batch_file}')

        batch = PBS(
            batch_file,
            log_file,
            cmds,
            walltime,
            memreq,
            rungroup=instance_settings['main_rungroup'],
            xnat_host=xnat_host,
            job_template=job_template
        )

        # Save to file
        batch.write()

        # Submit the saved file
        jobid, job_failed = batch.submit()
        logger.info(f'job submit results:{jobid}:{job_failed}')

        return jobid, label

    def get_cmds(self, host, user, input_list, analysis):
        # Get inputs text
        inputs = self.build_inputs_text(input_list)

        # Get main commands text
        maincmd = self.build_main_text(analysis)

        # Append other paths
        cmd = CMDS_TEMPLATE.format(
            inputs=inputs,
            version=analysis['procversion'],
            host=host,
            user=user,
            project=analysis['project'],
            analysis=analysis['label'],
            repo=analysis['procrepo'],
            maincmd=maincmd
        )

        return [cmd]

    def build_inputs_text(self, inputs):
        txt = '(\n'

        for cur in inputs:
            cur['fpath'] = requests.utils.quote(cur['fpath'], safe=":/")
            txt += '{fdest},{ftype},{fpath},{ddest}\n'.format(**cur)

        txt += ')\n\n'

        return txt

    def build_dir_text(self, analysis):
        # Set shared jobdir
        txt = 'JOBDIR=$(mktemp -d "{}.XXXXXXXXX") || '.format(self._rcqdir)
        txt += '{ echo "mktemp failed"; exit 1; }\n'
        txt += 'INDIR=$JOBDIR/INPUTS\n'
        txt += 'OUTDIR=$JOBDIR/OUTPUTS\n'

        return txt

    def build_main_text(self, analysis):
        pre = ''
        post = ''
        txt = ''

        if analysis['processor'].get('jobdir', '') == 'shared':
            txt += self.build_dir_text(analysis)

        # Get the pre command that runs before main
        if analysis['processor'].get('pre', False):
            pre = self.build_command(analysis, 'pre') + ' && '

        # Get the post command that runs after main
        if analysis['processor'].get('post', False):
            post = ' && ' + self.build_command(analysis, 'post')

        # Build and append the commands
        txt += 'MAINCMD="'
        txt += pre + self.build_command(analysis, 'command') + post

        # Finish with a newline
        txt += '\"\n'

        # Return the whole command lines
        return txt

    def build_command(self, analysis, name):
        txt = ''

        command = analysis['processor'][name]

        # Use the container name to get the path
        command['container'] = self.get_container_path(
            analysis['processor']['containers'],
            command['container'])

        # Build and append the post command
        if 'type' not in command:
            err = 'command type not set'
            logger.error(err)
            raise Exception(err)

        if analysis['processor'].get('jobdir', '') == 'shared':
            shared = True
        else:
            shared = False

        if command['type'] == 'singularity_run':
            txt = self.build_singularity_cmd('run', command, shared)

        elif command['type'] == 'singularity_exec':
            txt = self.build_singularity_cmd('exec', command, shared)

        else:
            err = 'invalid command type: {}'.format(command['type'])
            logger.error(err)
            raise Exception(err)

        return txt

    def build_singularity_cmd(self, runexec, command, shared=False):

        if 'container' not in command:
            err = 'singularity modes require a container to be set'
            logger.error(err)
            raise Exception(err)

        if runexec not in ['run', 'exec']:
            err = f'singularity mode {runexec} not known'
            logger.error(err)
            raise Exception(err)

        # Initialize command
        if 'opts' in command:
            # Get the user defined opts
            _opts = command['opts']
            # Prepend clean and contain
            _opts = f'--contain --cleanenv {_opts}'
            command_txt = f'singularity {runexec} {_opts}'
        else:
            command_txt = f'singularity {runexec} {SINGULARITY_BASEOPTS}'

        if shared:
            _user, _host = get_this_instance().split('@')
            command_txt += f' -e --env USER={_user} --env HOSTNAME={_host} '
            command_txt += ' -B $HOME/.ssh:$HOME/.ssh'

        # Append extra options
        _extra = command.get('extraopts', None)
        if _extra:
            command_txt = '{} {}'.format(command_txt, _extra)

        # Append container name
        command_txt = '{} {}'.format(command_txt, command['container'])

        # Append arguments for the singularity entrypoint
        cargs = command.get('args', None)
        if cargs:
            # Unescape and then escape double quotes
            cargs = cargs.replace('\\"', '"').replace('"', '\\\"')
            command_txt = '{} {}'.format(command_txt, cargs)

        return command_txt

    def get_container_path(self, containers, name):
        cpath = None

        # Find the matching container
        for c in containers:
            if c['name'] == name:
                cpath = c['path']
                break

        return cpath


def get_updates(analysis):
    """Update information from local SLURM."""
    project = analysis['project']
    instance = analysis['redcap_repeat_instance']
    updates = {}

    if 'analysis_jobid' not in analysis:
        logger.info(f'no jobid for analysis:{project}:{instance}')
        return

    jobid = analysis['analysis_jobid']
    cmd = f'sacct -j {jobid}.batch --units G --noheader -p --format MaxRss,cputime,NodeList,Start,End,State'

    logger.debug(f'running command:{cmd}')

    try:
        output = subprocess.check_output(cmd, shell=True)
        output = output.decode().strip()
    except subprocess.CalledProcessError:
        logger.info(f'error running command:{cmd}')
        return

    if not output:
        logger.debug(f'no output')
        return

    logger.debug(output)

    job_mem,job_time,job_node,job_start,job_end,job_state,_ = output.split('|')

    # only update usage if job is not still running
    if job_state == 'RUNNING':
        logger.debug(f'job still running:{project}:{instance}')
    else:
        if job_mem and job_mem != analysis.get('analysis_memused', ''):
            updates['analysis_memused'] = job_mem

        if job_time and job_time != analysis.get('analysis_timeused', ''):
            updates['analysis_timeused'] = job_time

    if job_node and job_node != analysis.get('analysis_jobnode', ''):
        updates['analysis_jobnode'] = job_node

    if job_start and job_start != analysis.get('analysis_jobstarttime', ''):
        updates['analysis_jobstarttime'] = job_start

    if job_end and job_end != analysis.get('analysis_jobendtime', ''):
        updates['analysis_jobendtime'] = job_end

    if job_state != analysis['analysis_status']:
        logger.debug(f'changing status to:{job_state}')
        updates['analysis_status'] = job_state

    return updates




import logging
LOGGER = logging.getLogger('dax')

# TODO: BenM/asr_of_asr/should detect / report cycles
# TODO: BenM/asr_of_asr/needs proper unittesting


class ProcessorGraph:

    @staticmethod
    def processor_inputs_from_sources(yaml_sources):
        sources = dict()
        for name, source in yaml_sources:
            xnat = source.contents.get('inputs', {}).get('xnat', {})

            asrs = xnat.get('assessors', {})
            inputs = set()
            for a in asrs:
                types = a.get('types', '').split(',')

                inputs = inputs.union(types)
            sources[name] = list(inputs)
        return sources

    @staticmethod
    def get_forward_edges(nodes_to_src_edges):
        sink_edges = dict()
        # calculate a list of nodes to sink edges
        for k, v in list(nodes_to_src_edges.items()):
            if k not in sink_edges:
                sink_edges[k] = []
            for src in v:
                if src not in sink_edges:
                    sink_edges[src] = []
                sink_edges[src].append(k)
        for v in sink_edges:
            sink_edges[v] = sorted(sink_edges[v])

        return sink_edges

    @staticmethod
    def order_processors(processors, log=None):
        """
        Order a list of processors in dependency order.
        This method takes a list of processors and orders them so that:
        . if processor type b takes an input of processor type a, processor a
          appears in the list before processor b
        . processors without a well-formed name are placed at the end of the
          list in no particular order


        :param processors:
        :return:
        """
        processor_map = dict([(p.get_proctype(), p) for p in processors])
        named_processors = []
        unnamed_processors = []
        assessor_inputs = {}
        for p in processors:
            proctype = p.get_proctype()
            if not proctype:
                unnamed_processors.append(p)
            else:
                assessor_inputs[p.get_proctype()] =\
                    p.get_assessor_input_types()

        ordered_names = ProcessorGraph.order_from_inputs(assessor_inputs, log)
        for n in ordered_names:
            named_processors.append(processor_map[n])
        return named_processors + unnamed_processors

    # TODO: BenM/assessor_of_assessor/ideally, this would operate on a list of
    # processors rather than a list of yaml_sources, but this would require a
    # refactor of existing processors
    @staticmethod
    def order_from_inputs(artefacts_to_inputs, log=None):
        # artefact_type_map is a list of artefacts to their *inputs*.
        # consider the following graph:
        # a --> b --> d
        #   \     \     \
        #    \     \     \
        #     > c --> e --> f
        # this function takes a mapping from a node to its inputs:
        # a -> [], b -> [a], c -> [a], d -> [b], e -> [b, c], f -> [d, e]
        #
        # the algorithm then proceeds as follows:
        # 1. calculate a list of nodes to sink edges (fwd_edges)
        # 2. calculate in-degrees for the nodes, adding any node with no
        #    inputs to the satisfied node list
        # 3. keep taking nodes from the satisfied node list, following their
        #    sink edges and lowering the in-degree of each sink node by 1.
        #    Nodes that hit zero in-degree are added to the satisfied list and
        #    the current node is added to the ordered list
        open_nodes = dict()
        satisfied = list()
        ordered = list()
        fwd_edges = ProcessorGraph.get_forward_edges(artefacts_to_inputs)

        # calculate the 'in-degree' (number of incoming edges) for each
        # node; this is more easily done using the mapping of nodes to inputs.
        # if a node has an in-degree of zero then it can go onto the satisfied
        # list
        for k, v in list(artefacts_to_inputs.items()):
            open_nodes[k] = len(v)
            if len(v) == 0:
                satisfied.append(k)

        # keep picking nodes from the front of the satisfied list until there
        # are no more. each follow the sink edges for that node and reduce the
        # sink node in-degrees by 1. Any node whose in-degree falls to 0 is
        # added to the end of the satisfied list
        while len(satisfied) > 0:
            cur = satisfied[0]
            ordered.append(cur)
            satisfied = satisfied[1:]
            for sink in fwd_edges[cur]:
                open_nodes[sink] -= 1
                if open_nodes[sink] == 0:
                    satisfied.append(sink)

        unordered = list()
        for k, v in list(open_nodes.items()):
            if v > 0:
                unordered.append(k)
        unordered = sorted(unordered)

        if len(unordered) > 0 and log is not None:
            log.warning('Unable to order all processors:')
            log.warning('  Unordered: ' + ', '.join(unordered))
            regions = ProcessorGraph.tarjan(fwd_edges)
            if len(regions) < len(artefacts_to_inputs):
                log.warning('Cyclic processor dependencies detected:')
            regions.reverse()
            for r in [x for x in regions if len(x) > 1]:
                log.warning('  Cycle: ' + ', '.join(r))

        return ordered + unordered

    @staticmethod
    def tarjan(graph):

        class Vertex:
            def __init__(self, v, e):
                self.v = v
                self.e = e
                self.index = None
                self.onstack = False
                self.lowLink = None

        class TarjanImpl:
            """
            Expecting graph to be in the form of:
            {
                a: [b, c],
                b: [d],
                c: [e],
                d: [f],
                e: [f],
                f: []
            }
            :param graph:
            :return:
            """

            def __init__(self):
                self.V = dict()
                self.S = list()
                self.index = 0
                self.scrs = list()

            def go(self, graph):
                self.V = dict()
                self.S = list()
                self.index = 0
                self.scrs = list()
                self.V = {v: Vertex(v, w) for (v, w) in list(graph.items())}

                self.index = 0
                self.S = list()

                for v in self.V.values():
                    if v.index is None:
                        self.strongconnect(v)

                return self.scrs

            def strongconnect(self, v):
                v.index = self.index
                v.lowlink = self.index
                self.index += 1
                self.S.append(v)
                v.onstack = True

                for d in v.e:
                    w = self.V[d]
                    if w.index is None:
                        self.strongconnect(w)
                        v.lowlink = min(v.lowlink, w.lowlink)
                    elif w.onstack is True:
                        v.lowlink = min(v.lowlink, w.index)

                if v.lowlink == v.index:
                    scr = []
                    while True:
                        w = self.S.pop()
                        w.onstack = False
                        scr.append(w.v)
                        if w is v:
                            break
                    self.scrs.append(scr)

        t = TarjanImpl()
        results = t.go(graph)
        # print results, len(results) < len(graph)
        return results



""" Processor class define for Scan and Session."""

import logging
import re
import os
import json
import itertools
import requests

from uuid import uuid4
from datetime import date

from . import XnatUtils, task
from . import assessor_utils
from . import processor_parser
from . import yaml_doc
from .errors import AutoProcessorError
from .dax_settings import DEFAULT_FS_DATATYPE, DEFAULT_DATATYPE
from .task import NeedInputsException
from .processors_v3 import Processor_v3, SgpProcessor


__copyright__ = 'Copyright 2013 Vanderbilt University. All Rights Reserved'
__all__ = ['Processor', 'AutoProcessor']

# Logger for logs
LOGGER = logging.getLogger('dax')


class Processor(object):
    """ Base class for processor """
    def __init__(self, walltime_str, memreq_mb, spider_path,
                 version=None, ppn=1, env=None, suffix_proc='',
                 xsitype='proc:genProcData',
                 job_template=None):
        """
        Entry point of the Base class for processor.

        :param walltime_str: Amount of walltime to request for the process
        :param memreq_mb: Number of megabytes of memory to use
        :param spider_path: Fully qualified path to the spider to run
        :param version: Version of the spider
        :param ppn: Number of processors per job to use.
        :param env: Environment file to source.
        :param suffix_proc: Processor suffix (if desired)
        :param xsitype: the XNAT xsiType.
        :return: None

        """
        self.job_template = job_template
        self.walltime_str = walltime_str  # 00:00:00 format
        self.memreq_mb = memreq_mb   # memory required in megabytes
        # default values:
        self.version = "1.0.0"
        # Suffix
        if suffix_proc and suffix_proc[0] != '_':
            self.suffix_proc = '_%s' % suffix_proc
        else:
            self.suffix_proc = suffix_proc
        self.suffix_proc = re.sub('[^a-zA-Z0-9]', '_', self.suffix_proc)
        self.name = None
        self.spider_path = spider_path
        self.ppn = ppn
        if env:
            self.env = env
        else:
            self.env = os.path.join(os.environ['HOME'], '.bashrc')
        self.xsitype = xsitype
        # getting name and version from spider_path
        self.set_spider_settings(spider_path, version)
        # if suffix_proc is empty, set it to "" for the spider call:
        if not suffix_proc:
            self.suffix_proc = ''

    # get the spider_path right with the version:
    def set_spider_settings(self, spider_path, version):
        """
        Method to set the spider version, path, and name from filepath

        :param spider_path: Fully qualified path and file of the spider
        :param version: version of the spider
        :return: None

        """
        if version:
            # get the proc_name
            proc_name = os.path.basename(spider_path)[7:-3]
            # remove any version if there is one
            proc_name = re.split("/*_v[0-9]/*", proc_name)[0]
            # setting the version and name of the spider
            self.version = version
            nformat = '''{procname}_v{version}{suffix}'''
            self.name = nformat.format(procname=proc_name,
                                       version=self.version.split('.')[0],
                                       suffix=self.suffix_proc)
            sformat = '''Spider_{procname}_v{version}.py'''
            spider_name = sformat.format(procname=proc_name,
                                         version=version.replace('.', '_'))
            self.spider_path = os.path.join(os.path.dirname(spider_path),
                                            spider_name)
        else:
            self.default_settings_spider(spider_path)

    def default_settings_spider(self, spider_path):
        """
        Get the default spider version and name

        :param spider_path: Fully qualified path and file of the spider
        :return: None

        """
        # set spider path
        self.spider_path = spider_path
        # set the name and the version of the spider
        if len(re.split("/*_v[0-9]/*", spider_path)) > 1:
            basename = os.path.basename(spider_path)
            self.version = basename[7:-3].split('_v')[-1].replace('_', '.')
            spidername = basename[7:-3]
            self.name = '''{procname}_v{version}{suffix}'''.format(
                procname=re.split("/*_v[0-9]/*", spidername)[0],
                version=self.version.split('.')[0], suffix=self.suffix_proc)
        else:
            self.name = os.path.basename(spider_path)[7:-3] + self.suffix_proc

    def get_proctype(self):
        """
        Return the processor name for this processor. Override this method if
        you are inheriting from a non-yaml processor.
        :return: the name of the processor type
        """

        return None

    def get_assessor_input_types(self):
        """
        Enumerate the assessor input types for this. The default implementation
        returns an empty collection; override this method if you are inheriting
        from a non-yaml processor.
        :return: a list of input assessor types
        """
        return []

    # should_run - is the object of the proper object type?
    # e.g. is it a scan? and is it the required scan type?
    # e.g. is it a T1?
    # other arguments here, could be Proj/Subj/Sess/Scan/Assessor depending
    # on processor type?
    def should_run(self):
        """
        Responsible for determining if the assessor should shouw up in session.

        :raises: NotImplementedError if not overridden.
        :return: None

        """
        raise NotImplementedError()

    def build_cmds(self, cobj, dir):

        """
        Build the commands that will go in the PBS/SLURM script
        :raises: NotImplementedError if not overridden from base class.
        :return: None
        """
        raise NotImplementedError()

    def create_assessor(self, xnatsession, inputs, relabel=False):
        attempts = 0
        while attempts < 100:
            guid = str(uuid4())
            assessor = xnatsession.assessor(guid)
            if not assessor.exists():
                kwargs = {}
                if self.xsitype.lower() == DEFAULT_FS_DATATYPE.lower():
                    fsversion = '{}/fsversion'.format(self.xsitype.lower())
                    kwargs[fsversion] = 0
                elif self.xsitype.lower() == DEFAULT_DATATYPE.lower():
                    proctype = '{}/proctype'.format(self.xsitype.lower())
                    kwargs[proctype] = self.name
                    procversion = '{}/procversion'.format(self.xsitype.lower())
                    kwargs[procversion] = self.version
                input_key = '{}/inputs'.format(self.xsitype.lower())
                kwargs[input_key] = self._serialize_inputs(inputs)
                if relabel:
                    _proj = assessor.parent().parent().parent().label()
                    _subj = assessor.parent().parent().label()
                    _sess = assessor.parent().label()
                    label = '-x-'.join([_proj, _subj, _sess, self.name, guid])
                else:
                    label = guid

                # Set creation date to today
                date_key = '{}/date'.format(self.xsitype.lower())
                date_val = str(date.today())
                kwargs[date_key] = date_val

                # Create the assessor
                assessor.create(assessors=self.xsitype.lower(),
                                ID=guid, label=label,
                                **kwargs)
                return assessor

            attempts += 1

    def _serialize_inputs(self, inputs):
        return json.dumps(inputs)

    def _deserialize_inputs(self, assessor):
        return json.loads(
            XnatUtils.parse_assessor_inputs(assessor.attrs.get('inputs')))


class AutoProcessor(Processor):
    """ Auto Processor class for AutoSpider using YAML files"""
    def __init__(self, xnat, yaml_source, user_inputs=None):
        """
        Entry point for the auto processor
        :param xnat: xnat context object (XnatUtils in production contexts)
        :param yaml_source: dictionary containing source_type -> string,\
                            source_id -> string, document -> yaml document
        :param user_inputs: a dictionary of user overrides to the yaml\
                            source document
        :return: None

        """
        if not xnat:
            raise AutoProcessorError("Parameter 'xnat' must be provided")
        if not yaml_source:
            raise AutoProcessorError(
                "Parameter 'yaml_source' must be provided")

        self.xnat = xnat
        self.user_overrides = dict()
        self.extra_user_overrides = dict()

        self._read_yaml(yaml_source)

        # Edit the values from user inputs:
        if user_inputs is not None:
            self._edit_inputs(user_inputs, yaml_source)

        self.parser = processor_parser.ProcessorParser(
            yaml_source.contents, self.proctype)

        # Set up attrs:
        self.walltime_str = self.attrs.get('walltime')
        self.memreq_mb = self.attrs.get('memory')
        self.ppn = self.attrs.get('ppn', 1)
        self.env = self.attrs.get('env', None)
        self.xsitype = self.attrs.get('xsitype', 'proc:genProcData')
        self.full_regex = self.attrs.get('fullregex', False)
        self.suffix = self.attrs.get('suffix', None)

    def _edit_inputs(self, user_inputs, yaml_source):
        """
        Method to edit the inputs from the YAML file by the user inputs.

        :param user_inputs: dictionary of tag, value. E.G:
            user_inputs = {'default.spider_path': /.../Spider....py'}
        """
        yaml_name = yaml_source.source_id

        for key, val in list(user_inputs.items()):
            LOGGER.debug('overriding:key={}, file={}'.format(key, yaml_name))
            tags = key.split('.')
            if key.startswith('inputs.default'):
                # change value in inputs
                if tags[-1] in list(self.user_overrides.keys()):
                    self.user_overrides[tags[-1]] = val
                elif tags[-1] in list(self.extra_user_overrides.keys()):
                    self.extra_user_overrides[tags[-1]] = val
                else:
                    msg = 'key not found in default inputs:key={}, file={}'
                    msg = msg.format(tags[-1], yaml_name)
                    LOGGER.error(msg)
                    raise AutoProcessorError(msg)

            elif key.startswith('inputs.xnat'):
                # change value in self.xnat_inputs
                if tags[2] not in list(self.xnat_inputs.keys()):
                    msg = 'key not found in xnat inputs:key={}, file={}'
                    msg = msg.format(tags[3], yaml_name)
                    LOGGER.error(msg)
                    raise AutoProcessorError(msg)

                # Match the scan number or assessor number (e.g: scan1)
                sobj = None
                for obj in self.xnat_inputs[tags[2]]:
                    if tags[3] == obj['name']:
                        sobj = obj
                        break

                if sobj is None:
                    msg = 'invalid override:key={}, file={}'
                    msg = msg.format(key, yaml_name)
                    LOGGER.error(msg)
                    raise AutoProcessorError(msg)

                if tags[4] == 'resources':
                    if tags[6] == 'fmatch':
                        # Match the resource name
                        robj = None
                        for obj in sobj['resources']:
                            if tags[5] == obj['varname']:
                                robj = obj
                                break

                        if robj is None:
                            msg = 'invalid override:key={}, file={}'
                            LOGGER.error(msg)
                            raise AutoProcessorError(msg)

                        msg = 'overriding fmatch:key={}, val={}'
                        msg = msg.format(key, val)
                        LOGGER.debug(msg)
                        robj['fmatch'] = val
                    else:
                        msg = 'invalid override:key={}, file={}'
                        msg = msg.format(key, yaml_name)
                        LOGGER.error(msg)
                        raise AutoProcessorError(msg)
                else:
                    LOGGER.info('overriding:{}:{}'.format(tags[4], str(val)))
                    obj[tags[4]] = val

            elif key.startswith('attrs'):
                # change value in self.attrs
                if tags[-1] in list(self.attrs.keys()):
                    self.attrs[tags[-1]] = val
                else:
                    msg = 'key not found in attrs:key={}, file={}'
                    msg = msg.format(tags[-1], yaml_name)
                    LOGGER.error(msg)
                    raise AutoProcessorError(msg)

            else:
                msg = 'invalid override:key={}, file={}'
                msg = msg.format(key, yaml_name)
                LOGGER.error(msg)
                raise AutoProcessorError(msg)

    def _read_yaml(self, yaml_source):
        """
        Method to parse the processor arguments and their default values.

        :param yaml_source: YamlDoc object containing the yaml file contents
        """
        if yaml_source.source_type is None:
            raise AutoProcessorError('Empty yaml source provided')

        doc = yaml_source.contents

        # Set Inputs from Yaml
        self._check_default_keys(yaml_source.source_id, doc)
        self.attrs = doc.get('attrs')
        self.command = doc.get('command')
        inputs = doc.get('inputs')
        self.xnat_inputs = inputs.get('xnat')
        for key, value in list(inputs.get('default').items()):
            # If value is a key in command
            k_str = '{{{}}}'.format(key)
            if k_str in self.command:
                self.user_overrides[key] = value
            else:
                if isinstance(value, bool) and value is True:
                    self.extra_user_overrides[key] = ''
                elif value and value != 'None':
                    self.extra_user_overrides[key] = value

        # Getting proctype from Yaml
        self.proctype, self.version = self.xnat.get_proctype(
            self.user_overrides.get('spider_path'),
            self.attrs.get('suffix', None))

        # Set attributes:
        self.spider_path = self.user_overrides.get('spider_path')
        self.name = self.proctype

        # Override template
        if doc.get('jobtemplate'):
            _tmp = doc.get('jobtemplate')

            # Make sure we have the full path
            if not os.path.isabs(_tmp):
                # If only filename, we assume it is same folder as default
                _tmp = os.path.join(os.path.dirname(self.job_template), _tmp)

            # Override it
            self.job_template = os.path.join(_tmp)

    def _check_default_keys(self, source_id, doc):
        """ Static method to raise error if key not found in dictionary from
        yaml file.

        :param source_id: dictionary containing source_type -> string,\
                            source_id -> string, document -> yaml document
        :param key: key to check in the doc
        """
        # first level
        for key in ['inputs', 'command', 'attrs']:
            self._raise_yaml_error_if_no_key(doc, source_id, key)
        # Second level in inputs and attrs:
        inputs = doc.get('inputs')
        attrs = doc.get('attrs')
        for _doc, key in [(inputs, 'default'), (inputs, 'xnat'),
                          (attrs, 'memory'),
                          (attrs, 'walltime')]:
            self._raise_yaml_error_if_no_key(_doc, source_id, key)
        # third level for default:
        default = doc.get('inputs').get('default')
        for key in ['spider_path']:
            self._raise_yaml_error_if_no_key(default, source_id, key)

    @ staticmethod
    def _raise_yaml_error_if_no_key(doc, source_id, key):
        """Method to raise an execption if the key is not in the dict

        :param doc: dict to check
        :param source_id: YAML source identifier string for logging
        :param key: key to search
        """
        if key not in list(doc.keys()):
            err = 'YAML source {} does not have {} defined. See example.'
            raise AutoProcessorError(err.format(source_id, key))

    def parse_session(self, csess, sessions, pets=None):
        """
        Method to run the processor parser on this session, in order to
        calculate the pattern matches for this processor and the sessions
        provided
        :param csess: the active session. For non-longitudinal studies, this is
        the session that the pattern matching is performed on. For longitudinal
        studies, this is the 'current' session from which all prior sessions
        are numbered for the purposes of pattern matching
        :param sessions: the full, time-ordered list of sessions that should be
        considered for longitudinal studies.
        :return: None
        """
        return self.parser.parse_session(csess, sessions, pets)

    def get_proctype(self):
        return self.name
    # ***** Names still need fixing! *****

    def get_assessor_input_types(self):
        """
        Enumerate the assessor input types for this. The default implementation
        returns an empty collection; override this method if you are inheriting
        from a non-yaml processor.
        :return: a list of input assessor types
        """
        assessor_inputs = [i for i in list(self.parser.inputs.values()) if i['artefact_type'] == 'assessor']
        assessors = [i['types'] for i in assessor_inputs]

        return list(itertools.chain.from_iterable(assessors))

    # TODO: BenM/assessor_of_assessor/this method is no longer suitable for
    # execution on a single assessor, as it generates commands for the whole
    # session. In any case, the command should be written out to the xnat
    # assessor schema so that it can simply be launched, rather than performing
    # this reconstruction each time the dax launch command is called
    def get_cmds(self, assr, jobdir):
        """Method to generate the spider command for cluster job.

        :param assessor: pyxnat assessor object
        :param jobdir: jobdir where the job's output will be generated
        :return: command to execute the spider in the job script
        """

        # TODO: BenM/assessor_of_assessors/parse each scan / assessor and
        # any select statements and generate one or more corresponding commands

        # self.parser.generate_command(cassr, )

        # combine the user overrides with the input parameters for each
        # distinct command
        commands = []
        variable_set = self.parser.get_variable_set(assr)
        combined_params = {}
        for k, v in list(variable_set.items()):
            combined_params[k] = v
        for k, v in list(self.user_overrides.items()):
            combined_params[k] = v

        cmd = self.command.format(**combined_params)

        for key, value in list(self.extra_user_overrides.items()):
            cmd = '{} --{} {}'.format(cmd, key, value)

        # TODO: BenM/assessor_of_assessor/each assessor is separate and
        # has a different label; change the code to fetch the label from
        # the assessor
        # Add assr and jobdir:
        assr_full_name =\
            assessor_utils.full_label_from_assessor(assr)
        if ' -a ' not in cmd and ' --assessor ' not in cmd:
            cmd = '{} -a {}'.format(cmd, assr_full_name)
        if ' -d ' not in cmd:
            cmd = '{} -d {}'.format(cmd, jobdir)

        commands.append(cmd)

        return commands


class MoreAutoProcessor(AutoProcessor):
    """ More Auto Processor class for AutoSpider using YAML files"""

    def __init__(self, xnat, yaml_source, user_inputs=None,
                 singularity_imagedir=None, job_template='~/job_template.txt'):

        """
        Entry point for the auto processor

        :param yaml_file: yaml file defining the processor
        :return: None

        """

        # Load outputs
        self.outputs = dict()

        # Save location of singularity imagedir
        self.singularity_imagedir = singularity_imagedir

        # Set the template to the global default, it could be overwritten by
        # processor yaml
        self.job_template = job_template

        super(MoreAutoProcessor, self).__init__(xnat, yaml_source, user_inputs)

    def _read_yaml(self, yaml_source):
        """
        Method to read the processor arguments and there default value.

        :param yaml_file: path to yaml file defining the processor
        """
        if yaml_source.source_type is None:
            raise AutoProcessorError('Empty yaml source provided')

        doc = yaml_source.contents

        # Set Inputs from Yaml
        self._check_default_keys(yaml_source.source_id, doc)
        self.attrs = doc.get('attrs')
        self.command = doc.get('command')

        # Set Inputs from Yaml
        inputs = doc.get('inputs')
        self.xnat_inputs = inputs.get('xnat')
        for key, value in list(inputs.get('default').items()):
            # If value is a key in command
            k_str = '{{{}}}'.format(key)
            if k_str in self.command:
                self.user_overrides[key] = value
            else:
                if isinstance(value, bool) and value is True:
                    self.extra_user_overrides[key] = ''
                elif value and value != 'None':
                    self.extra_user_overrides[key] = value

        # Container path, prepend singularity imagedir
        self.container_path = inputs.get('default').get('container_path')
        if ((self.container_path.endswith('.simg') or
                self.container_path.endswith('.img') or
                self.container_path.endswith('.sif')) and
                not os.path.isabs(self.container_path) and
                self.singularity_imagedir):
            self.container_path = os.path.join(
                self.singularity_imagedir, self.container_path)

        # Overwrite container_path for building script
        self.user_overrides['container_path'] = self.container_path

        # Getting proctype and version from Yaml
        if doc.get('procversion'):
            self.version = doc.get('procversion')
        else:
            self.version = self.parse_procversion()

        if doc.get('procname'):
            procname = doc.get('procname')
        else:
            procname = self.parse_procname()

        if doc.get('proctype'):
            self.proctype = doc.get('proctype')
        else:
            self.proctype = '{}_v{}'.format(
                procname, self.version.split('.')[0]
            )

        suffix = self.attrs.get('suffix', None)
        if suffix:
            if suffix[0] != '_':
                suffix = '_{}'.format(suffix)

            suffix = re.sub('[^a-zA-Z0-9]', '_', suffix)

            if suffix[-1] == '_':
                suffix = suffix[:-1]

            self.proctype = '{}{}'.format(self.proctype, suffix)

        # Set attributes:
        self.name = self.proctype

        # Set Outputs from Yaml
        self.outputs = doc.get('outputs')

        # Override template
        if doc.get('jobtemplate'):
            _tmp = doc.get('jobtemplate')

            # Make sure we have the full path
            if not os.path.isabs(_tmp):
                # If only filename, we assume it is same folder as default
                _tmp = os.path.join(os.path.dirname(self.job_template), _tmp)

            # Override it
            self.job_template = os.path.join(_tmp)

    def _check_default_keys(self, source_id, doc):
        """ Static method to raise error if key not found in dictionary from
        yaml file.

        :param yaml_file: path to yaml file defining the processor
        :param doc: doc dictionary extracted from the yaml file
        :param key: key to check in the doc
        """
        # first level
        for key in ['inputs', 'command', 'attrs', 'outputs']:
            self._raise_yaml_error_if_no_key(doc, source_id, key)

        # Second level in inputs and attrs:
        inputs = doc.get('inputs')
        attrs = doc.get('attrs')
        for _doc, key in [(inputs, 'default'), (inputs, 'xnat'),
                          (attrs, 'memory'),
                          (attrs, 'walltime')]:
            self._raise_yaml_error_if_no_key(_doc, source_id, key)

        # third level for default:
        default = doc.get('inputs').get('default')

        for key in ['container_path']:
            self._raise_yaml_error_if_no_key(default, source_id, key)

    def parse_procname(self):
        tmp = self.container_path
        tmp = tmp.split('://')[-1]
        tmp = tmp.rsplit('/')[-1]

        if len(re.split('/*[_:]v[0-9]/*', tmp)) > 1:
            tmp = re.split('/*[_:]v[0-9]/*', tmp)[0]

        if tmp.startswith('Spider_'):
            tmp = tmp[len('Spider_'):]

        return tmp

    def parse_procversion(self):
        tmp = self.container_path

        tmp = tmp.split('://')[-1]
        tmp = tmp.rsplit('/')[-1]

        if tmp.endswith('.img'):
            tmp = tmp.split('.img')[0]
        elif tmp.endswith('.simg'):
            tmp = tmp.split('.simg')[0]
        elif tmp.endswith('.py'):
            tmp = tmp.split('.py')[0]

        if len(re.split('/*_v[0-9]/*', tmp)) > 1:
            tmp = tmp.split('_v')[-1].replace('_', '.')
        elif len(re.split('/*:v[0-9]/*', tmp)) > 1:
            tmp = tmp.split(':v')[-1].replace('_', '.')

        return tmp

    def build_cmds(self, assr, assr_label, sessions, jobdir, resdir):
        """Method to generate the spider command for cluster job.
        :param jobdir: jobdir where the job's output will be generated
        :return: command to execute the spider in the job script
        """
        assr_inputs = XnatUtils.get_assessor_inputs(assr, sessions)

        # Make every input a list, so we can iterate later
        for k in assr_inputs:
            if not isinstance(assr_inputs[k], list):
                assr_inputs[k] = [assr_inputs[k]]

        # Find values for the xnat inputs
        var2val, input_list = self.parser.find_inputs(
            assr, sessions, assr_inputs)

        # Append other stuff
        for k, v in list(self.user_overrides.items()):
            var2val[k] = v

        for k, v in list(self.extra_user_overrides.items()):
            var2val[k] = v

        # Include the assessor label
        var2val['assessor'] = assr_label

        # Handle xnat attributes
        for attr_in in self.xnat_inputs.get('attrs', list()):
            _var = attr_in['varname']
            _attr = attr_in['attr']
            _obj = attr_in['object']
            _val = ''

            if _obj == 'subject':
                _val = assr.parent().parent().attrs.get(_attr)
            elif _obj == 'session':
                _val = assr.parent().attrs.get(_attr)
            elif _obj == 'scan':
                _ref = attr_in['ref']
                _refval = [a.rsplit('/', 1)[1] for a in assr_inputs[_ref]]
                _val = ','.join(
                    [assr.parent().scan(r).attrs.get(_attr) for r in _refval])
            elif _obj == 'assessor':
                if 'ref' in attr_in:
                    _ref = attr_in['ref']
                    _refval = [a.rsplit('/', 1)[1] for a in assr_inputs[_ref]]
                    _val = ','.join(
                        [assr.parent().assessor(r).attrs.get(_attr) for r in _refval])
                else:
                    _val = assr.attrs.get(_attr)
            else:
                LOGGER.error('invalid YAML')
                err = 'YAML File:contains invalid attribute:{}'
                raise AutoProcessorError(err.format(_attr))

            if _val == '':
                raise NeedInputsException('Missing ' + _attr)
            else:
                var2val[_var] = _val

        # Handle edits
        edit_res = assr.out_resource(task.EDITS_RESOURCE)
        if edit_res.exists():
            file_list = edit_res.files().get()
            assr_path = '/projects/{}/subjects/{}/experiments/{}/assessors/{}'.format(
                assr.parent().parent().parent().label(),
                assr.parent().parent().label(),
                assr.parent().label(),
                assr.label()
                )

            for edit_in in self.xnat_inputs.get('edits', list()):
                _fpref = edit_in['fpref']
                _var = edit_in['varname']

                # Filter files that match prefix
                cur_list = [f for f in file_list if f.startswith(_fpref)]

                if cur_list:
                    # Sort and grab the last file
                    _val = sorted(cur_list)[-1]

                    # Build full uri
                    _uri = '{}/data{}/out/resources/{}/files/{}'.format(
                        assr._intf.host,
                        assr_path,
                        task.EDITS_RESOURCE,
                        _val)

                    # Append to inputs to be downloaded
                    input_list.append({
                        'fdest': _fpref,
                        'ftype': 'FILE',
                        'fpath': _uri,
                        'ddest': ''
                    })

                    # Set the value for command text
                    var2val[_var] = '/INPUTS/'+_fpref

                else:
                    # None found
                    var2val[_var] = ''
        else:
            for edit_in in self.xnat_inputs.get('edits', list()):
                var2val[edit_in['varname']] = ''

        # Build the command text
        dstdir = os.path.join(resdir, assr_label)
        assr_dir = os.path.join(jobdir, assr_label)
        cmd = self.build_text(
            var2val, input_list, assr_dir, dstdir,
            assr._intf.host, assr._intf.user)

        return [cmd]

    def build_text(self, var2val, input_list, jobdir, dstdir, host, user):
        # Initialize commands
        cmd = '\n\n'

        # Append the list of inputs, URL-encoding the fpath to handle special chars in URLs
        cmd += 'INLIST=(\n'
        for cur in input_list:
            cur['fpath'] = requests.utils.quote(cur['fpath'],safe=":/")
            cmd += '{fdest},{ftype},{fpath},{ddest}\n'.format(**cur)

        cmd += ')\n\n'

        # Append the list on outputs
        cmd += 'OUTLIST=(\n'
        for cur in self.outputs:
            cmd += '{path},{type},{resource}\n'.format(**cur)

        cmd += ')\n\n'

        # Append other paths
        cmd += 'VERSION={}\n'.format(self.version)
        cmd += 'JOBDIR=$(mktemp -d "{}.XXXXXXXXX") || '.format(jobdir)
        cmd += '{ echo "mktemp failed"; exit 1; }\n'
        cmd += 'INDIR=$JOBDIR/INPUTS\n'
        cmd += 'OUTDIR=$JOBDIR/OUTPUTS\n'
        cmd += 'DSTDIR={}\n\n'.format(dstdir)
        cmd += 'CONTAINERPATH={}\n\n'.format(self.container_path)
        cmd += 'XNATHOST={}\n\n'.format(host)
        cmd += 'XNATUSER={}\n\n'.format(user)

        # Append the main command
        cmd += 'MAINCMD=\"'
        cmd += self.command.format(**var2val)
        cmd += '\"\n'

        return cmd


def processors_by_type(proc_list):
    """
    Organize the processor types and return a list of session processors
     first, then scan

    :param proc_list: List of Processor classes from the DAX settings file
    :return: Lists of processors by type

    """
    auto_proc_list = list()

    # Build list of processors by type
    if proc_list is not None:
        for proc in proc_list:
            if isinstance(proc, Processor_v3):
                auto_proc_list.append(proc)
            elif issubclass(proc.__class__, AutoProcessor):
                auto_proc_list.append(proc)
            else:
                LOGGER.warn('unknown processor type: %s' % proc)

    return auto_proc_list


def load_from_yaml(xnat, filepath, user_inputs=None,
                   singularity_imagedir=None, job_template=None):
    """
    Load processor from yaml
    :param filepath: path to yaml file
    :return: processor
    """
    yaml_obj = yaml_doc.YamlDoc().from_file(filepath)

    # Load file based on yaml version and data type
    if yaml_obj.contents.get('inputs').get('xnat').get('sessions', False):
        # This must be a subjgenproc
        return SgpProcessor(
            xnat, filepath, user_inputs, singularity_imagedir, job_template)
    elif yaml_obj.contents.get('procyamlversion', '') == '3.0.0-dev.0':
        LOGGER.debug('loading as Processor_v3:{}'.format(filepath))
        return Processor_v3(
            xnat, filepath, user_inputs, singularity_imagedir, job_template)
    elif yaml_obj.contents.get('moreauto'):
        return MoreAutoProcessor(xnat, yaml_obj, user_inputs,
                                 singularity_imagedir, job_template)
    else:
        return AutoProcessor(xnat, yaml_obj, user_inputs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" launcher.py that represents the main object called by the executables """

from datetime import datetime, timedelta
import logging
import sys
import os
import traceback
import tempfile
import time

from . import lockfiles
from . import processors, modules, XnatUtils, task, cluster, processors_v3
from .task import Task, ClusterTask, XnatTask, mkdirp
from .dax_settings import DAX_Settings, DAX_Netrc
from .errors import (ClusterCountJobsException, ClusterLaunchException,
                     DaxXnatError, DaxLauncherError)
from . import yaml_doc
from .processor_graph import ProcessorGraph
from .utilities import find_with_pred, groupby_to_dict

__copyright__ = 'Copyright 2019 Vanderbilt University. All Rights Reserved'
__all__ = ['Launcher']
DAX_SETTINGS = DAX_Settings()
UPDATE_PREFIX = 'updated--'
UPDATE_FORMAT = "%Y-%m-%d %H:%M:%S"
BUILD_SUFFIX = 'BUILD_RUNNING.txt'
UPDATE_SUFFIX = 'UPDATE_RUNNING.txt'
LAUNCH_SUFFIX = 'LAUNCHER_RUNNING.txt'

# Logger to print logs
LOGGER = logging.getLogger('dax')


def str_to_timedelta(delta_str):
    if len(delta_str) <= 1:
        raise ValueError('invalid timedelta string value')

    val = int(delta_str[:-1])
    if delta_str.endswith('s'):
        return timedelta(seconds=val)
    elif delta_str.endswith('m'):
        return timedelta(minutes=val)
    elif delta_str.endswith('h'):
        return timedelta(hours=val)
    elif delta_str.endswith('d'):
        return timedelta(days=val)
    else:
        raise ValueError('invalid timedelta string value')


def check_res_dir(resdir):
    check_dir(os.path.join(resdir, 'DISKQ'))
    check_dir(os.path.join(os.path.join(resdir, 'DISKQ'), 'INPUTS'))
    check_dir(os.path.join(os.path.join(resdir, 'DISKQ'), 'OUTLOG'))
    check_dir(os.path.join(os.path.join(resdir, 'DISKQ'), 'BATCH'))
    check_dir(os.path.join(resdir, 'FlagFiles'))


def check_dir(dir_path):
    try:
        os.makedirs(dir_path)
    except OSError:
        if not os.path.isdir(dir_path):
            raise


def task_needs_to_run(procstatus, qcstatus):
    return\
        procstatus in [task.NEED_TO_RUN, task.NEED_INPUTS] or\
        qcstatus in [task.RERUN, task.REPROC, task.DOES_NOT_EXIST]


def task_needs_status_update(qcstatus):
    return qcstatus in [task.RERUN, task.REPROC]


class Launcher(object):
    """ Launcher object to manage a list of projects from a settings file """

    def __init__(self,
                 resdir,
                 project_process_dict=dict(),
                 project_modules_dict=dict(),
                 yaml_dict=dict(),
                 priority_project=None,
                 queue_limit=10,
                 queue_limit_pending=10,
                 limit_pendinguploads=10,
                 launch_delay_sec=1,
                 root_job_dir='/tmp',
                 xnat_user=None, xnat_pass=None, xnat_host=None, cr=None,
                 job_email=None, job_email_options='FAIL', job_rungroup=None,
                 launcher_type='diskq-combined',
                 job_template='~/job_template.txt',
                 smtp_host=None,
                 timeout_emails=None,
                 project_sgp_processors={}):
        """
        Entry point for the Launcher class

        :param project_process_dict: dictionary associating project & processes
        :param project_modules_dict: dictionary associating project & modules
        :param priority_project: list of project to describe the priority
        :param queue_limit: maximum number of jobs in the queue
        :param queue_limit_pending: maximum pending jobs in the queue
        :param limit_pendinguploads: maximum number of uploads waiting
        :param launch_delay_sec: time to wait between job launches
        :param root_job_dir: root directory for jobs
        :param xnat_host: XNAT Host url. By default, use env variable.
        :param cr: True if the host is an XNAT CR instance (will default to
        :False if not specified)
        :param xnat_user: XNAT User ID. By default, use env variable.
        :param xnat_pass: XNAT Password. By default, use env variable.
        :param job_email: job email address for report
        :param job_email_options: email options for the jobs
        :param job_rungroup: cluster group to run the job under
        :param max_age: maximum time before updating again a session
        :return: None
        """
        self.queue_limit = queue_limit
        self.queue_limit_pending = queue_limit_pending
        self.limit_pendinguploads = limit_pendinguploads
        self.launch_delay_sec = launch_delay_sec
        self.root_job_dir = root_job_dir
        self.resdir = resdir
        self.smtp_host = smtp_host
        self.timeout_emails = timeout_emails

        # Processors:
        if not isinstance(project_process_dict, dict):
            err = 'project_process_dict set but it is not a dictionary with \
project name as a key and list of processors objects as values.'
            raise DaxLauncherError(err)
        self.project_process_dict = (project_process_dict
                                     if project_process_dict is not None
                                     else dict())

        # Modules:
        if not isinstance(project_modules_dict, dict):
            err = 'project_modules_dict set but it is not a dictionary with \
project name as a key and list of processors objects as values.'
            raise DaxLauncherError(err)
        self.project_modules_dict = (project_modules_dict
                                     if project_modules_dict is not None
                                     else dict())

        # YAML dict:
        if not isinstance(yaml_dict, dict):
            err = 'Yaml_files set but it is not a dictionary with project \
name as a key and list of yaml filepaths as values.'
            raise DaxLauncherError(err)

        # TODO: should this self be here, I don't think so - bdb 2022-04-05
        self.yaml_dict = yaml_dict if yaml_dict is not None else dict()

        # Add processors to project_process_dict:
        for project, yaml_objs in list(yaml_dict.items()):
            for yaml_obj in yaml_objs:
                if isinstance(yaml_obj, processors_v3.Processor_v3):
                    proc = yaml_obj
                elif isinstance(yaml_obj, processors.AutoProcessor):
                    proc = yaml_obj
                elif isinstance(yaml_obj, str):
                    # TODO: BenM/general_refactor/this logic should be handled
                    # further up the call stack - launchers should be provided
                    # AutoProcessors rather than strings for yaml files
                    yaml_obj = yaml_doc.YamlDoc().from_file(yaml_obj)
                    proc = processors.AutoProcessor(XnatUtils, yaml_obj)
                elif isinstance(yaml_obj.yaml_doc.YamlDoc):
                    proc = processors.AutoProcessor(XnatUtils, yaml_obj)
                else:
                    err = 'yaml_obj of type {} is unsupported'
                    raise DaxLauncherError(err.format(type(yaml_obj)))

                if project not in self.project_process_dict:
                    self.project_process_dict[project] = [proc]
                else:
                    self.project_process_dict[project].append(proc)

        self.project_sgp_processors = project_sgp_processors

        if isinstance(priority_project, list):
            self.priority_project = priority_project
        elif isinstance(priority_project, str):
            self.priority_project = priority_project.split(',')
        else:
            self.priority_project = None

        self.job_email = job_email
        self.job_email_options = job_email_options
        self.job_rungroup = job_rungroup
        self.launcher_type = launcher_type
        self.job_template = job_template

        # Create Folders for flagfile/pbs/outlog in RESULTS_DIR
        check_res_dir(resdir)

        self.xnat_host = xnat_host
        if not self.xnat_host:
            self.xnat_host = os.environ['XNAT_HOST']

        # CR flag: don't want something like 'cr: blah blah' in the settings
        # file turning the cr flag on
        if str(cr).upper() == 'TRUE':
            self.cr = True
        else:
            self.cr = False

        LOGGER.info(
            'XNAT CR status: cr=%s, self.cr=%s' % (str(cr), str(self.cr)))

        # User:
        if not xnat_user:
            netrc_obj = DAX_Netrc()
            user, password = netrc_obj.get_login(self.xnat_host)
            self.xnat_user = user
            self.xnat_pass = password
        else:
            self.xnat_user = xnat_user
            if not xnat_pass:
                msg = 'Please provide password for host <%s> and user <%s>: '
                self.xnat_pass = eval(
                    input(msg % (self.xnat_host, self.xnat_user)))
            else:
                self.xnat_pass = xnat_pass

    # LAUNCH Main Method
    def launch_jobs(self, lockfile_prefix, project_local, sessions_local,
                    writeonly=False, pbsdir=None, force_no_qsub=False):
        """
        Main Method to launch the tasks

        :param lockfile_prefix: prefix for flag file to lock the launcher
        :param project_local: project to run locally
        :param sessions_local: list of sessions to launch tasks
         associated to the project locally
        :param writeonly: write the job files without submitting them
        :param pbsdir: folder to store the pbs file
        :param force_no_qsub: run the job locally on the computer (serial mode)
        :return: None

        """
        if self.launcher_type == 'diskq-xnat':
            err = 'cannot launch jobs with this launcher type: %s'
            raise DaxLauncherError(err % self.launcher_type)

        LOGGER.info('-------------- Launch Tasks --------------')
        LOGGER.info('launcher_type = %s' % self.launcher_type)

        flagfile = os.path.join(os.path.join(self.resdir, 'FlagFiles'),
                                '%s_%s' % (lockfile_prefix, LAUNCH_SUFFIX))

        project_list = self.init_script(flagfile, project_local,
                                        type_update=3, start_end=1)

        if project_list is None or len(project_list) == 0:
            LOGGER.info('no projects to launch')
        else:
            msg = 'Loading task queue from: %s'
            LOGGER.info(msg % os.path.join(self.resdir, 'DISKQ'))
            task_list = load_task_queue(
                self.resdir,
                status=task.NEED_TO_RUN,
                proj_filter=project_list,
                sess_filter=sessions_local)

            msg = '%s tasks that need to be launched found'
            LOGGER.info(msg % str(len(task_list)))
            self.launch_tasks(task_list, force_no_qsub=force_no_qsub)

        self.finish_script(flagfile, project_list, 3, 2, project_local)

    @staticmethod
    def is_launchable_tasks(assr_info):
        """
        Check if a task is launchable

        :param assr_info: dictionary containing procstatus for the assessor
        :return: True if tasks need to be launch, False otherwise.
        """
        return assr_info['procstatus'] == task.NEED_TO_RUN

    def launch_tasks(self, task_list, force_no_qsub=False):
        """
        Launch tasks from the passed list until the queue is full or
         the list is empty

        :param task_list: list of task to launch
        :param force_no_qsub: run the job locally on the computer (serial mode)
        :return: None
        """
        launched, pending, pendinguploads = cluster.count_jobs(self.resdir,force_no_qsub)
        if not force_no_qsub:
            LOGGER.info(
                'Cluster: %d/%d total, %d/%d pending, %d/%d pending uploads',
                launched, self.queue_limit,
                pending, self.queue_limit_pending,
                pendinguploads, self.limit_pendinguploads
                )

        # Launch until we reach cluster limit or no jobs left to launch
        _launchcnt = 0
        _launchmax = len(task_list)
        while(
                launched < self.queue_limit and
                pending < self.queue_limit_pending and
                pendinguploads < self.limit_pendinguploads and
                len(task_list) > 0
                ):

            cur_task = task_list.pop()

            LOGGER.info('Launching job: %s', cur_task.assessor_label)

            try:
                success = cur_task.launch(force_no_qsub=force_no_qsub)
            except Exception as E:
                LOGGER.critical(
                    'Caught exception launching job %s',
                    cur_task.assessor_label
                    )
                LOGGER.critical(
                    'Exception class %s caught with message %s',
                    E.__class__, str(E)
                    )
                LOGGER.critical(traceback.format_exc())
                success = False

            if not success:
                LOGGER.error('ERROR: failed to launch job')
                raise ClusterLaunchException

            _launchcnt = _launchcnt + 1
            time.sleep(self.launch_delay_sec)

            launched, pending, pendinguploads = cluster.count_jobs(self.resdir,force_no_qsub)
            if not force_no_qsub:
                LOGGER.info(
                    'Cluster: %d/%d total, %d/%d pending, %d/%d pending uploads',
                    launched, self.queue_limit,
                    pending, self.queue_limit_pending,
                    pendinguploads, self.limit_pendinguploads
                    )

        if not force_no_qsub:
            LOGGER.info(
                'Launched %d of %d jobs. Stopping', _launchcnt, _launchmax
                )


    def update_tasks(self, lockfile_prefix, project_local, sessions_local):
        """
        Main method to Update the tasks

        :param lockfile_prefix: prefix for flag file to lock the launcher
        :param project_local: project to run locally
        :param sessions_local: list of sessions to update tasks associated
         to the project locally
        :return: None

        """
        if self.launcher_type == 'diskq-xnat':
            err = 'cannot update jobs with this launcher type: %s'
            raise DaxLauncherError(err % self.launcher_type)

        LOGGER.info('-------------- Update Tasks --------------')
        LOGGER.info('launcher_type = %s' % self.launcher_type)

        flagfile = os.path.join(os.path.join(self.resdir, 'FlagFiles'),
                                '%s_%s' % (lockfile_prefix, UPDATE_SUFFIX))
        project_list = self.init_script(flagfile, project_local,
                                        type_update=2, start_end=1)

        if project_list is None or len(project_list) == 0:
            LOGGER.info('no projects to launch')
        else:
            msg = 'Loading task queue from: %s'
            LOGGER.info(msg % os.path.join(self.resdir, 'DISKQ'))
            task_list = load_task_queue(self.resdir, proj_filter=project_list)

            LOGGER.info('%s tasks found.' % str(len(task_list)))

            LOGGER.info('Updating tasks...')
            for cur_task in task_list:
                LOGGER.info('Updating task: %s' % cur_task.assessor_label)
                cur_task.update_status()

        self.finish_script(flagfile, project_list, 2, 2, project_local)

    @staticmethod
    def is_updatable_tasks(assr_info):
        """
        Check if a task is updatable.

        :param assr_info: dictionary containing procstatus/qcstatus
        :return: True if tasks need to be update, False otherwise.

        """
        good_proc = assr_info['procstatus'] in task.OPEN_STATUS_LIST
        good_qc = assr_info['qcstatus'] in task.OPEN_QA_LIST
        return good_proc or good_qc

    # BUILD Main Method
    def build(self, lockfile_prefix, project_local, sessions_local,
              mod_delta=None, proj_lastrun=None, start_sess=None):
        """
        Main method to build the tasks and the sessions

        :param lockfile_prefix: prefix for flag file to lock the launcher
        :param project_local: project to run locally
        :param sessions_local: list of sessions to launch tasks
         associated to the project locally
        :return: None

        """
        if self.launcher_type == 'diskq-cluster':
            err = 'cannot build jobs with this launcher type: %s'
            raise DaxLauncherError(err % self.launcher_type)

        LOGGER.info('-------------- Build --------------')
        LOGGER.info('launcher_type = %s' % self.launcher_type)
        LOGGER.info('mod delta = %s' % str(mod_delta))

        flagfile = os.path.join(os.path.join(self.resdir, 'FlagFiles'),
                                '%s_%s' % (lockfile_prefix, BUILD_SUFFIX))
        project_list = self.init_script(flagfile, project_local,
                                        type_update=1, start_end=1)

        LOGGER.info('Connecting to XNAT at %s' % self.xnat_host)
        with XnatUtils.get_interface(
            self.xnat_host,
            self.xnat_user,
            self.xnat_pass,
            self.smtp_host,
            self.timeout_emails
        ) as intf:

            if not XnatUtils.has_dax_datatypes(intf):
                err = 'error: dax datatypes are not installed on xnat <%s>'
                raise DaxXnatError(err % (self.xnat_host))

            # Priority if set:
            if self.priority_project and not project_local:
                unique_list = set(list(self.project_process_dict.keys()) +
                    list(self.project_modules_dict.keys()) +
                    list(self.project_sgp_processors.keys()))
                project_list = self.get_project_list(list(unique_list))

            # Build projects
            for project_id in project_list:
                LOGGER.info('===== PROJECT: %s =====' % project_id)
                try:
                    if ((proj_lastrun) and
                            (project_id in proj_lastrun) and
                            (proj_lastrun[project_id] is not None)):
                        lastrun = proj_lastrun[project_id]
                    else:
                        lastrun = None

                    self.build_project(intf, project_id, lockfile_prefix,
                                       sessions_local,
                                       mod_delta=mod_delta, lastrun=lastrun,
                                       start_sess=start_sess)

                    if len(self.get_subjgenproc_processors(project_id)) > 0:
                        # The project has SGP processors so we build them
                        self.build_project_subjgenproc(intf, project_id)

                except Exception as E:
                    err1 = 'Caught exception building project %s'
                    err2 = 'Exception class %s caught with message %s'
                    LOGGER.critical(err1 % project_id)
                    LOGGER.critical(err2 % (E.__class__, str(E)))
                    LOGGER.critical(traceback.format_exc())

        self.finish_script(flagfile, project_list, 1, 2, project_local)

    def build_project(self, intf, project_id, lockfile_prefix, sessions_local,
                      mod_delta=None, lastrun=None, start_sess=None):
        """
        Build the project

        :param intf: pyxnat.Interface object
        :param project_id: project ID on XNAT
        :param lockfile_prefix: prefix for flag file to lock the launcher
        :param sessions_local: list of sessions to launch tasks
        :return: None
        """

        # TODO: make a project settings to store, processors,
        #       modules, etc

        # Get lists of modules/processors per scan/exp for this project
        proj_mods = self.project_modules_dict.get(project_id, None)
        proj_procs = self.project_process_dict.get(project_id, None)
        exp_mods, scan_mods = modules.modules_by_type(proj_mods)
        auto_procs = processors.processors_by_type(proj_procs)
        auto_procs = ProcessorGraph.order_processors(auto_procs, LOGGER)

        if proj_mods:
            # Modules prerun
            LOGGER.info('* Modules Prerun')
            if sessions_local:
                self.module_prerun(project_id, 'manual_update')
            else:
                self.module_prerun(project_id, lockfile_prefix)

        if mod_delta:
            lastmod_delta = str_to_timedelta(mod_delta)
        else:
            lastmod_delta = None

        # get the list of processors for this project
        processor_types = set([x.name for x in auto_procs])

        LOGGER.info('* Loading list of sessions from XNAT for project')
        sess_list = self.get_sessions_list(intf, project_id, sessions_local)

        # Skip to session
        if start_sess:
            for i, sess in enumerate(sess_list):
                if sess['label'] == start_sess:
                    LOGGER.info('starting index=' + str(i))
                    sess_list = sess_list[i:]
                    break

        # Group by subject
        sessions_by_subject = groupby_to_dict(
            sess_list, lambda x: x['subject_id'])

        # check for processor types that are new to this project
        assr_types = intf.list_project_assessor_types(project_id)
        has_new = (len(processor_types.difference(assr_types)) > 0)
        LOGGER.debug(assr_types)
        LOGGER.debug('has_new=' + str(has_new))

        for subject_id, sessions in list(sessions_by_subject.items()):
            # Get the cached session objects for this subject

            sessions_to_update = dict()

            # Check which sessions (if any) require an update:
            for sess_info in sessions:
                if has_new:
                    # Don't skip any sessions
                    pass
                elif lastrun:
                    last_mod = datetime.strptime(
                        sess_info['last_modified'][0:19], UPDATE_FORMAT)

                    if last_mod < lastrun:
                        mess = "+ Session %s:skipping not modified since last run, last_mod=%s, last_run=%s"
                        LOGGER.info(mess % (sess_info['label'], str(last_mod),
                                            str(lastrun)))
                        continue

                elif lastmod_delta:
                    last_mod = datetime.strptime(
                        sess_info['last_modified'][0:19], UPDATE_FORMAT)
                    now_date = datetime.today()
                    if now_date > last_mod + lastmod_delta:
                        mess = "+ Session %s:skipping not modified within delta, last_mod=%s"
                        LOGGER.info(mess % (sess_info['label'], str(last_mod)))
                        continue
                    else:
                        LOGGER.info('+ Session {}:modified, last_mod={}'.format(
                            sess_info['label'], str(last_mod)))

                # Append session to list of sessions to update
                sessions_to_update[sess_info['ID']] = sess_info

            if len(sessions_to_update) == 0:
                continue

            # build a full list of sessions for the subject: they may be needed
            # even if not all sessions are getting updated
            mess = "+ Subject %s: loading XML for %s session(s)..."
            LOGGER.info(mess % (sessions[0]['subject_label'], len(sessions)))
            cached_sessions = [XnatUtils.CachedImageSession(
                intf, x['project_label'], x['subject_label'],
                x['session_label']) for x in sessions]

            if len(cached_sessions) > 1:
                cached_sessions = sorted(
                    cached_sessions,
                    key=lambda s: s.creation_timestamp(), reverse=True)

            # update each of the sessions that require it
            for sess_info in list(sessions_to_update.values()):
                try:
                    # TODO: BenM - ensure that this code is robust to subjects
                    # without sessions and sessions without assessors / scans
                    mess = "+ Session %s: building..."
                    LOGGER.info(mess % sess_info['label'])
                    self.build_session(
                        intf, sess_info, auto_procs, exp_mods, scan_mods,
                        sessions=cached_sessions)
                except Exception as E:
                    err1 = 'Caught exception building sessions %s'
                    err2 = 'Exception class %s caught with message %s'
                    LOGGER.critical(err1 % sess_info['session_label'])
                    LOGGER.critical(err2 % (E.__class__, str(E)))
                    LOGGER.critical(traceback.format_exc())

        if not sessions_local or sessions_local.lower() == 'all':
            # Modules after run
            LOGGER.debug('* Modules Afterrun')
            try:
                self.module_afterrun(intf, project_id)
            except Exception as E:
                err2 = 'Exception class %s caught with message %s'
                LOGGER.critical('Caught exception after running modules')
                LOGGER.critical(err2 % (E.__class__, str(E)))
                LOGGER.critical(traceback.format_exc())

    def build_project_subjgenproc(self, xnat, project, includesubj=None):
        """
            Build the project

            :param xnat: pyxnat.Interface object
            :param project: project ID on XNAT
            :param includesubj: specific subjects to build, otherwise all
            :return: None
        """
        LOGFORMAT = '%(asctime)s:%(levelname)s:%(module)s:%(message)s'
        stdouthandler = logging.StreamHandler(sys.stdout)
        stdouthandler.setFormatter(logging.Formatter(LOGFORMAT))
        LOGGER.addHandler(stdouthandler)
        LOGGER.setLevel(logging.DEBUG)

        pdata = {}
        pdata['name'] = project
        pdata['scans'] = XnatUtils.load_scan_data(xnat, [project])
        pdata['assessors'] = XnatUtils.load_assr_data(xnat, [project])
        pdata['sgp'] = XnatUtils.load_sgp_data(xnat, project)

        LOGGER.debug('calling build_sgp_processors')
        self.build_sgp_processors(xnat, pdata, includesubj)

    def build_sgp_processors(self, xnat, project_data, includesubj=None):
        project = project_data['name']
        sgp_processors = self.get_subjgenproc_processors(project)
        subjects = project_data['sgp'].SUBJECT.unique()

        # Filter list if specified
        if includesubj:
            LOGGER.debug('no subjects specified, including all')
            subjects = [x for x in subjects if x in includesubj]

        if len(sgp_processors) == 0:
            LOGGER.debug('no sgp processors')
            return

        for subj in sorted(subjects):
            for processor in sgp_processors:
                # Get list of inputs sets (not yet matched with existing)
                inputsets = processor.parse_subject(subj, project_data)

                for inputs in inputsets:
                    if inputs == {}:
                        # print('empty set, skipping')
                        return

                    # Get(create) assessor with given inputs and proc type
                    # TODO: extract subject data only
                    (assr, info) = processor.get_assessor(
                        xnat, subj, inputs, project_data)

                    # TODO: apply reproc or rerun if needed
                    # (assr,info) = undo_processing()
                    # (assr,info) = reproc_processing()

                    if info['PROCSTATUS'] in [task.NEED_TO_RUN, task.NEED_INPUTS]:
                        # print('building task')
                        (assr, info) = self.build_task(
                            assr, info, processor, project_data)

                        # print('assr after=', info)
                    else:
                        LOGGER.info('already built:{}'.format(info['ASSR']))

    def build_task(self, assr, info, processor, project_data):
        resdir = self.resdir
        old_proc_status = info['PROCSTATUS']
        old_qc_status = info['QCSTATUS']
        jobdir = self.root_job_dir

        try:
            cmds = processor.build_cmds(
                assr,
                info,
                project_data,
                jobdir,
                resdir)

            batch_file = self.batch_path(info['ASSR'])
            outlog = self.outlog_path(info['ASSR'])

            batch = cluster.PBS(
                batch_file,
                outlog,
                cmds,
                processor.walltime_str,
                processor.memreq_mb,
                processor.ppn,
                processor.env,
                self.job_email,
                self.job_email_options,
                self.job_rungroup,
                self.xnat_host,
                processor.job_template)

            LOGGER.info('writing:' + batch_file)
            batch.write()

            # Set new statuses to be updated
            new_proc_status = task.JOB_RUNNING
            new_qc_status = task.JOB_PENDING

            # Write processor spec file for version 3
            try:
                LOGGER.debug('writing processor spec file')
                filename = self.processor_spec_path(info['ASSR'])
                mkdirp(os.path.dirname(filename))
                processor.write_processor_spec(filename)
            except AttributeError as err:
                # older processor does not have version
                LOGGER.debug('procyamlversion not found'.format(err))

        except task.NeedInputsException as e:
            new_proc_status = task.NEED_INPUTS
            new_qc_status = e.value
        except task.NoDataException as e:
            new_proc_status = task.NO_DATA
            new_qc_status = e.value

        # Update on xnat
        if new_proc_status != old_proc_status:
            assr.attrs.set(
                'proc:subjgenprocdata/procstatus', new_proc_status)

        if new_qc_status != old_qc_status:
            assr.attrs.set(
                'proc:subjgenprocdata/validation/status', new_qc_status)

        # Update local info
        info['PROCSTATUS'] = new_proc_status
        info['QCSTATUS'] = new_qc_status

        return (assr, info)

    # TODO:BenM/assessor_of_assessor/modify from here for one to many
    # processor to assessor mapping
    def build_session(self, intf, sess_info, auto_proc_list,
                      sess_mod_list, scan_mod_list,
                      sessions):
        """
        Build a session

        :param intf: pyxnat.Interface object
        :param sess_info: python dictionary from XnatUtils.list_sessions method
        :param auto_proc_list: list of processors running
        :param sess_mod_list: list of modules running on a session
        :param scan_mod_list: list of modules running on a scan
        :return: None
        """
        csess = find_with_pred(
            sessions, lambda s: sess_info['label'] == s.session_id())

        init_timestamp = csess.cached_timestamp

        # Create log file for this build of this session
        now_time = datetime.strftime(datetime.now(), '%Y%m%d-%H%M%S')
        sess_label = csess.label()
        tmp_dir = tempfile.mkdtemp()
        tmp_name = '{}_build_log-{}.txt'.format(sess_label, now_time)
        tmp_file = os.path.join(tmp_dir, tmp_name)
        handler = logging.FileHandler(tmp_file, 'w')
        handler.setFormatter(logging.Formatter(
            fmt='%(asctime)s - %(levelname)s - %(module)s - %(message)s'))
        LOGGER.addHandler(handler)

        if sess_mod_list or scan_mod_list:
            # Modules
            mod_count = 0
            while mod_count < 3:
                mess = """== Build modules (count:{count}) =="""
                LOGGER.debug(mess.format(count=mod_count))
                if sess_mod_list:
                    self.build_session_modules(intf, csess, sess_mod_list)
                if scan_mod_list:
                    for cscan in csess.scans():
                        LOGGER.debug('+SCAN: ' + cscan.info()['scan_id'])
                        self.build_scan_modules(intf, cscan, scan_mod_list)

                reloaded = csess.refresh()
                if not reloaded:
                    break

                mod_count += 1

        # Auto Processors
        if auto_proc_list:
            LOGGER.debug('== Build auto processors ==')
            self.build_auto_processors(csess, auto_proc_list, sessions)

        # Close sess log
        LOGGER.handlers.pop()

        # Upload build log only if session was changed
        csess.refresh()
        final_timestamp = csess.cached_timestamp
        LOGGER.debug('initial timestamp={}, final timestamp={}'.format(
            init_timestamp, final_timestamp))
        if final_timestamp > init_timestamp:
            res_obj = csess.full_object().resource('BUILD_LOGS')
            LOGGER.debug('uploading session log:' + tmp_file)
            XnatUtils.upload_file_to_obj(tmp_file, res_obj)
        else:
            LOGGER.debug('session not modified, not uploading build log')

    def build_session_modules(self, xnat, csess, sess_mod_list):
        """
        Build a session

        :param xnat: pyxnat.Interface object
        :param sess_info: python ditionary from XnatUtils.list_sessions method
        :param sess_mod_list: list of modules running on a session
        :return: None
        """
        sess_obj = None
        sess_info = csess.info()
        for sess_mod in sess_mod_list:
            LOGGER.debug('* Module: ' + sess_mod.getname())
            if sess_mod.needs_run(csess, xnat):
                if sess_obj is None:
                    sess_obj = csess.full_object()

                try:
                    sess_mod.run(sess_info, sess_obj)
                except Exception as E:
                    err1 = 'Caught exception building session module %s'
                    err2 = 'Exception class %s caught with message %s'
                    LOGGER.critical(err1 % sess_info['session_label'])
                    LOGGER.critical(err2 % (E.__class__, str(E)))
                    LOGGER.critical(traceback.format_exc())

    def build_scan_modules(self, xnat, cscan, scan_mod_list):
        """ Build Scan modules.

        :param xnat: pyxnat.Interface object
        :param csan: CachedObject for scan (XnatUtils)
        :param scan_mod_list: list of modules running on a scan
        :return: None
        """
        scan_info = cscan.info()
        scan_obj = None

        # Modules
        for scan_mod in scan_mod_list:
            LOGGER.debug('* Module: ' + scan_mod.getname())
            if scan_mod.needs_run(cscan, xnat):
                if scan_obj is None:
                    scan_obj = cscan.full_object()

                try:
                    scan_mod.run(scan_info, scan_obj)
                except Exception as E:
                    err1 = 'Caught exception building session scan module \
in session %s'
                    err2 = 'Exception class %s caught with message %s'
                    LOGGER.critical(err1 % scan_info['session_label'])
                    LOGGER.critical(err2 % (E.__class__, str(E)))
                    LOGGER.critical(traceback.format_exc())

    def build_auto_processors(self, csess, auto_proc_list, sessions):
        """ Build yaml-based processors.

        :param xnat: pyxnat.Interface object
        :param csess: CachedObject for Session (XnatUtils)
        :param auto_proc_list: list of yaml processors
        :return: None
        """
        # sess_info = csess.info()
        xnat_session = csess.full_object()

        for auto_proc in auto_proc_list:
            # return a mapping between the assessor input sets and existing
            # assessors that map to those input sets

            # BDB 6/5/21
            # For now, we will not process pet sessions here, but will
            # include them as additional params for mr sessions.
            if csess.datatype() == 'xnat:petSessionData':
                LOGGER.debug('pet session, skipping auto processors')
                continue
            else:
                pets = [x for x in sessions if x.datatype() == 'xnat:petSessionData']
                mrs = [x for x in sessions if x.datatype() != 'xnat:petSessionData']

            mapping = auto_proc.parse_session(csess, mrs, pets)

            if mapping is None:
                continue

            for inputs, p_assrs in mapping:
                if len(p_assrs) == 0:
                    assessor = auto_proc.create_assessor(
                        xnat_session, inputs, relabel=True)
                    assessors = [(
                        assessor, assessor.label(),
                        task.NEED_TO_RUN, task.DOES_NOT_EXIST)]

                    csess.refresh()
                else:
                    assessors = []
                    for p in p_assrs:
                        # BDB 6/5/21 is there ever more than one assessor here?
                        info = p.info()
                        procstatus = info['procstatus']
                        qcstatus = info['qcstatus']
                        assessors.append((
                            p.full_object(), p.label(), procstatus, qcstatus))

                for assessor in assessors:
                    procstatus = assessor[2]
                    qcstatus = assessor[3]
                    if task_needs_to_run(procstatus, qcstatus):
                        xtask = XnatTask(auto_proc, assessor[0], self.resdir,
                                         os.path.join(self.resdir, 'DISKQ'))

                        if task_needs_status_update(qcstatus):
                            xtask.update_status()

                        LOGGER.debug('building task: ' + xtask.assessor_label)
                        (proc_status, qc_status) = xtask.build_task(
                            assessor[0], sessions,
                            self.root_job_dir,
                            self.job_email,
                            self.job_email_options,
                            self.job_rungroup)
                        deg = 'proc_status=%s, qc_status=%s'
                        LOGGER.debug(deg % (proc_status, qc_status))

                        # Refresh the cached session since we just modified it
                        csess.refresh()
                    else:
                        # TODO: check that it actually exists in QUEUE
                        LOGGER.debug('already built: ' + assessor[1])

    def batch_path(self, assessor_label):
        return os.path.join(
            self.resdir,
            'DISKQ',
            'BATCH',
            '{}.slurm'.format(assessor_label))

    def outlog_path(self, assessor_label):
        return os.path.join(
            self.resdir,
            'DISKQ',
            'OUTLOG',
            '{}.txt'.format(assessor_label))

    def processor_spec_path(self, assessor):
        return os.path.join(self.resdir, 'DISKQ', 'processor', assessor)

    def get_subjgenproc_processors(self, project):
        # Get the processors for this project
        proc = self.project_sgp_processors.get(project, [])

        # Filter to only include subjgenproc
        # proc = [x for x in proc if x.xsitype == 'proc:subjgenprocdata']

        return proc

    def set_subjgenproc_processors(self, project, processors):
        # Get the processors for this project
        self.sgp_processors[project] = processors

    def module_prerun(self, project_id, settings_filename=''):
        """
        Run the module prerun method

        :param xnat: pyxnat.Interface object
        :param project_id: project ID on XNAT
        :param settings_filename: Settings file name for temp dir
        :return: None
        """
        for mod in self.project_modules_dict.get(project_id, list()):
            try:
                mod.prerun(settings_filename)
            except Exception as E:
                err1 = 'Caught exception in module prerun for project %s'
                err2 = 'Exception class %s caught with message %s'
                LOGGER.critical(err1 % project_id)
                LOGGER.critical(err2 % (E.__class__, str(E)))
                LOGGER.critical(traceback.format_exc())

    def module_afterrun(self, xnat, project_id):
        """
        Run the module afterrun method

        :param xnat: pyxnat.Interface object
        :param project_id: project ID on XNAT
        :return: None
        """
        for mod in self.project_modules_dict.get(project_id, list()):
            try:
                mod.afterrun(xnat, project_id)
            except Exception as E:
                err1 = 'Caught exception in module prerun for project %s'
                err2 = 'Exception class %s caught with message %s'
                LOGGER.critical(err1 % project_id)
                LOGGER.critical(err2 % (E.__class__, str(E)))
                LOGGER.critical(traceback.format_exc())

    # Generic Methods
    def init_script(self, flagfile, project_local, type_update, start_end):
        """
        Init script for any of the main methods: build/update/launch

        :param flagfile: flag file for the method to run
        :param project_local: project to run locally
        :param type_update: What type of process ran: dax_build (1),
         dax_update_tasks (2), dax_launch (3)
        :param start_end: starting timestamp (1) and ending timestamp (2)
        :return: None
        """
        # Get default project list for XNAT out of the module/process dict
        ulist = set(list(self.project_process_dict.keys()) +
                    list(self.project_sgp_processors.keys()) +
                    list(self.project_modules_dict.keys()))
        project_list = sorted(ulist)
        if project_local:
            if ',' in project_local:
                mess = """too much projects ID given to the option\
--project : {proj}. Only for one project."""
                mess_str = mess.format(proj=project_local)
                LOGGER.error(mess_str)
                exit(1)
            elif project_local in project_list:
                # Updating session for a specific project
                project_list = [project_local]
            else:
                mess = """failed to run locally on project {proj}.\
The project is not part of the settings."""
                mess_str = mess.format(proj=project_local)
                LOGGER.error(mess_str)
                exit(1)
        else:
            success = lockfiles.lock_flagfile(flagfile)
            if not success:
                LOGGER.warn('failed to get lock. Already running.')
                exit(1)

        return project_list

    def finish_script(self, flagfile, project_list, type_update,
                      start_end, project_local):
        """
        Finish script for any of the main methods: build/update/launch

        :param flagfile: flag file for the method to run
        :param project_list: List of projects that were updated by the method
        :param type_update: What type of process ran: dax_build (1),
         dax_update_tasks (2), dax_launch (3)
        :param start_end: starting timestamp (1) and ending timestamp (2)
        :param project_local: project to run locally
        :return: None
        """
        if not project_local:
            lockfiles.unlock_flagfile(flagfile)

    def get_tasks(self, xnat, is_valid_assessor, project_list=None,
                  sessions_local=None):
        """
        Get list of tasks for a projects list

        :param xnat: pyxnat.Interface object
        :param is_valid_assessor: method to validate the assessor
        :param project_list: List of projects to search tasks from
        :param sessions_local: list of sessions to update tasks associated
         to the project locally
        :return: list of tasks
        """
        task_list = list()

        if not project_list:
            projects = list(self.project_process_dict.keys())
            # Priority:
            if self.priority_project:
                project_list = self.get_project_list(projects)
            else:
                project_list = list(projects)

        # iterate projects
        for project_id in project_list:
            LOGGER.info('===== PROJECT:%s =====' % project_id)
            task_list.extend(self.get_project_tasks(xnat,
                                                    project_id,
                                                    sessions_local,
                                                    is_valid_assessor))

        return task_list

    def get_project_tasks(self, xnat, project_id, sessions_local,
                          is_valid_assessor):
        """
        Get list of tasks for a specific project where each task agrees
         the is_valid_assessor conditions

        :param xnat: pyxnat.Interface object
        :param project_id: project ID on XNAT
        :param sessions_local: list of sessions to update tasks associated
         to the project locally
        :param is_valid_assessor: method to validate the assessor
        :return: list of tasks
        """
        task_list = list()

        # Get lists of processors for this project
        pp_dict = self.project_process_dict.get(project_id, None)
        auto_procs = processors.processors_by_type(pp_dict)

        # Get lists of assessors for this project
        assr_list = self.get_assessors_list(xnat, project_id, sessions_local)

        # Match each assessor to a processor, get a task, and add to list
        for assr_info in assr_list:
            if is_valid_assessor(assr_info):
                cur_task = self.generate_task(xnat, assr_info, auto_procs)
                if cur_task:
                    task_list.append(cur_task)

        return task_list

    @staticmethod
    def match_proc(assr_info, auto_proc_list):
        """
        Check if an assessor is a match with the processors

        :param assr_info: dictionary containing the assessor info
                          (See XnatUtils.list_assessors)
        :param auto_proc_list: list of processors running
        :return: processor if found, None otherwise
        """
        # Look for a match in yaml processors
        for auto_proc in auto_proc_list:
            if auto_proc.xsitype == assr_info['xsiType'] and \
                    auto_proc.name == assr_info['proctype']:
                return auto_proc

        return None

    def generate_task(self, xnat, assr_info, auto_proc_list):
        """
        Generate a task for the assessor in the info

        :param xnat: pyxnat.Interface object
        :param assr_info: dictionary containing the assessor info
                          (See XnatUtils.list_assessors)
        :param auto_proc_list: list of yaml processors
        :return: task if processor and assessor match, None otherwise
        """
        task_proc = self.match_proc(assr_info, auto_proc_list)

        if task_proc is None:
            warn = 'no matching processor found: %s'
            LOGGER.warn(warn % assr_info['assessor_label'])
            return None
        else:
            # Get a new task with the matched processor
            assr = xnat.select_assessor(assr_info['project_id'],
                                        assr_info['subject_id'],
                                        assr_info['session_id'],
                                        assr_info['ID'])
            cur_task = Task(task_proc, assr, self.resdir)
            return cur_task

    @staticmethod
    def get_assessors_list(xnat, project_id, slocal):
        """
        Get the assessor list from XNAT and filter it if necessary

        :param xnat: pyxnat.Interface object
        :param project_id: project ID on XNAT
        :param slocal: session selected by user
        :return: list of assessors for a project
        """
        # Get lists of assessors for this project
        assr_list = xnat.list_project_assessors(project_id)

        # filter the assessors to the sessions given as parameters if given
        if slocal and slocal.lower() != 'all':
            # filter the list and keep the match between both list:
            val = slocal.split(',')
            assr_list = [x for x in assr_list if x['session_label'] in val]
            if not assr_list:
                warn = 'No processes on XNAT matched the sessions given: %s .'
                LOGGER.warn(warn % slocal)
                sys.exit(1)

        return assr_list

    @staticmethod
    def get_sessions_list(xnat, project_id, slocal):
        """
        Get the sessions list from XNAT and sort it.
         Move the new sessions to the front.

        :param xnat: pyxnat.Interface object
        :param project_id: project ID on XNAT
        :param slocal: session selected by user
        :return: list of sessions sorted for a project
        """
        list_sessions = xnat.get_sessions_minimal(project_id)

        if slocal and slocal.lower() != 'all':
            # filter the list and keep the match between both list:
            val = slocal.split(',')
            list_sessions = list(
                [x for x in list_sessions if x['label'] in val])
            if not list_sessions:
                warn = 'No session from XNAT matched the sessions given: %s .'
                LOGGER.warn(warn % slocal)

        # TODO: sort by last modified
        sorted_list = list_sessions
        new_sessions_label = [sess['label'] for sess in list_sessions]
        for session in list_sessions:
            if not session['label'] in new_sessions_label:
                sorted_list.append(session)

        return sorted_list

    def get_project_list(self, all_projects):
        """
        Get project list from the file priority + the other ones

        :param all_projects: list of all the projects in the settings file
        :return: list of project sorted to update
        """
        random_project = [x for x in all_projects
                          if x not in self.priority_project]
        return self.priority_project + random_project

    @staticmethod
    def has_new_processors(assessors, proc_types):
        """
        Method to check whether, given a list of assessors, there are processor
        types that are new relative to the list of assessors (the proc type
        doesn't appear in the set of assessor proc types).
        :param assessors: a list of assessors, typically from a session or
        subject
        :param proc_types: a set of processor types to check against assessors
        :return: Boolean indicating whether the proc_types set has proc types
        that aren't in the assessors list
        """
        assr_types = set(x['proctype'] for x in assessors)
        return len(proc_types.difference(assr_types)) > 0

# =============================================================================
def load_task_queue(resdir, status=None, proj_filter=None, sess_filter=None):
    """ Load the task queue for DiskQ"""
    task_list = list()
    diskq_dir = os.path.join(resdir, 'DISKQ')

    # TODO: handle subjgenproc assessors, conveniently it works implicitly, but
    # should also handle subject filters
    for t in os.listdir(os.path.join(diskq_dir, 'BATCH')):
        if proj_filter or sess_filter:
            assr = XnatUtils.AssessorHandler(t)
            if proj_filter and assr.get_project_id() not in proj_filter:
                LOGGER.debug('ignoring:' + t)
                continue

            if sess_filter and assr.get_session_label() not in sess_filter:
                LOGGER.debug('ignoring:' + t)
                continue

        LOGGER.debug('loading:' + t)
        task = ClusterTask(os.path.splitext(t)[0], resdir, diskq_dir)
        LOGGER.debug('status = ' + task.get_status())

        if not status or task.get_status() == status:
            LOGGER.debug('adding task to list:' + t)
            task_list.append(task)

    return task_list


def get_sess_lastmod(xnat, sess_info):
    """ Get the session last modified date."""
    xsi_type = sess_info['xsiType']
    sess_obj = xnat.select_experiment(sess_info['project_label'],
                                      sess_info['subject_label'],
                                      sess_info['session_label'])
    last_modified_xnat = sess_obj.attrs.get('%s/meta/last_modified' % xsi_type)
    last_mod = datetime.strptime(last_modified_xnat[0:19], '%Y-%m-%d %H:%M:%S')
    return last_mod


def sess_was_modified(xnat, sess_info, build_start_time):
    """
    Compare modified time with start time
    :param xnat: pyxnat.Interface object
    :param sess_info: dictionary of session information
    :param update_start_time: date when the update started
    :return: False if the session change and don't set the last update date,
             True otherwise
    """
    last_mod = get_sess_lastmod(xnat, sess_info)
    return (last_mod > build_start_time)




from unittest import TestCase

import io
import json
import yaml
import xml.etree.cElementTree as xmlet

from dax import XnatUtils
from dax.tests import unit_test_common_processor_yamls as yamls
from dax.yaml_doc import YamlDoc
from dax import AutoProcessor
from dax.tests.common_session_tools import SessionTools


# TODO: BenM/xnat refactor/ these sanity checks should be refactored into
# integration tests and performed on a test database that is explicitly
# setup / torn down within the scope of a test suite execution.
# Remove everything that doesn't make it into an integration test _before_
# the pull request back to vuiis/dax!

host = 'http://10.1.1.17'
proj_id = 'proj1'
subj_id = 'subj1'
sess_id = 'sess1'
scan_id = '1'
assr_id = 'asr1'

asrxsitype = 'proc:genProcData'
scanxsitype = 'xnat:mrScanData'

expstr = '/data/projects/{}/subjects/{}/experiments/{}'.format(
    proj_id, subj_id, sess_id)
scanstr = '/scans/{}'


class SanityChecks(TestCase):

    @staticmethod
    def __get_connection():
        return XnatUtils.get_interface(host=host)

    @staticmethod
    def __add_scan(name, sess, outputs, types="T1", quality="usable"):
        scn = sess.scan(name)
        if scn.exists():
            scn.delete()
            scn = sess.scan(name)
        kwargs = dict()
        kwargs[scanxsitype + '/type'] = types
        kwargs[scanxsitype + '/quality'] = quality
        scn.create(scans=scanxsitype, **kwargs)
        for k, v in outputs.items():
            for vv in v:
                scn.resource(k).file(vv[0]).insert(vv[1])
        return scn

    @staticmethod
    def __add_assessor(name, sess, inputs, outputs):
        asr = sess.assessor(name)
        if asr.exists():
            asr.delete()
            asr = sess.assessor(name)
        kwargs = dict()
        kwargs[asrxsitype + '/proctype'] = 'proc1'
        kwargs[asrxsitype + '/procversion'] = '1.0.0'
        kwargs[asrxsitype + '/validation/status'] = "Needs QA"
        kwargs[asrxsitype + '/inputs'] = json.dumps(inputs)
        asr.create(assessors=asrxsitype, **kwargs)
        for k, v in outputs.items():
            for vv in v:
                asr.resource(k).file(vv[0]).insert(vv[1])
        return asr

    @staticmethod
    def __prep_project(intf):
        sess = intf.select_experiment(proj_id, subj_id, sess_id)
        print((sess.exists()))
        scn1_outputs = {
            'NIFTI': [('images.nii', './file.txt')],
            'SNAPSHOTS': [
                ('snapshot.gif', './file.txt'),
                ('snapshot(1).gif', './file.txt')
            ]
        }
        scn1 = SanityChecks.__add_scan(
            '1', sess, scn1_outputs, quality="unusable")

        asr1_outputs = {
            'SEG': [('seg.nii.gz', './file.txt')]
        }
        asr1 = SanityChecks.__add_assessor('asr1', sess, {
            'scan1': expstr + scanstr.format('1')
        }, asr1_outputs)

    def test_interface_select_project(self):
        with SanityChecks.__get_connection() as intf:
            projobj = intf.select_project(proj_id)
            print(projobj)

    def test_interface_select_subject(self):
        with SanityChecks.__get_connection() as intf:
            subjobj = intf.select_subject(proj_id, subj_id)
            print(subjobj)

    def test_interface_select_session(self):
        with SanityChecks.__get_connection() as intf:
            sessobj = intf.select_experiment(proj_id, subj_id, sess_id)
            print(sessobj)

    def test_session_full_object(self):
        with SanityChecks.__get_connection() as intf:
            sessobj1 = intf.select_experiment(proj_id, subj_id, sess_id)
            sessobj2 = XnatUtils.CachedImageSession(
                intf, proj_id, subj_id, sess_id).full_object()
            self.assertEqual(sessobj1.label(), sessobj2.label())

    def test_scan_full_object(self):
        with SanityChecks.__get_connection() as intf:
            scanobj1 = intf.select_scan(proj_id, subj_id, sess_id, scan_id)
            csess = XnatUtils.CachedImageSession(
                intf, proj_id, subj_id, sess_id)
            cscan = [x for x in csess.scans() if x.label() == scan_id][0]
            scanobj2 = cscan.full_object()
            self.assertEqual(scanobj1.label(), scanobj2.label())

    def test_assessor_full_object(self):
        with SanityChecks.__get_connection() as intf:
            assrobj1 = intf.select_assessor(proj_id, subj_id, sess_id, assr_id)
            csess = XnatUtils.CachedImageSession(
                intf, proj_id, subj_id, sess_id)
            cassr = [x for x in csess.assessors() if x.label() == assr_id][0]
            assrobj2 = cassr.full_object()
            self.assertEqual(assrobj1.label(), assrobj2.label())

    def test_list_project_scans(self):
        with SanityChecks.__get_connection() as intf:
            print((intf.get_project_scans(intf, proj_id)))

    def test_processor_get_cmd(self):
        with SanityChecks.__get_connection() as intf:
            xnat = XnatUtils
            yaml_source = {
                'source_type': 'string',
                'source_id': 'unit test string',
                'document': yaml.load(
                    (io.StringIO(yamls.scan_gif_parcellation_yaml)))
            }
            ap = AutoProcessor(xnat, yaml_source)
            assr = intf.select_assessor(proj_id, subj_id, sess_id, assr_id)
            print((ap.get_cmds(assr, 'ajobdir')))

    def test_list_project_scans(self):
        with SanityChecks.__get_connection() as intf:
            scans = intf.get_project_scans(proj_id, False)
            print((len(scans)))

    def test_sanity_check_element_tree(self):
        xmltext = """
            <xnat:a xmlns:xnat="xnat">
                <xnat:out>
                    <xnat:file id='x'>
                        <nugget>Hi</nugget>
                    </xnat:file>
                    <xnat:file id='y'></xnat:file>
                </xnat:out>
                <xnat:out>
                    <xnat:file id='z'></xnat:file>
                    <xnat:file id='w'></xnat:file>
                </xnat:out>
             </xnat:a>
             """

        tree = xmlet.fromstring(xmltext)
        print(tree)
        print(('find "xnat:out":', tree.find('xnat:out', {'xnat': 'xnat'})))
        print(('find "xnat:file":', tree.find('xnat:file', {'xnat': 'xnat'})))
        print(('find "xnat:out->xnat:file":', tree.find(
                'xnat:out', {'xnat': 'xnat'}).find(
                'xnat:file', {'xnat': 'xnat'})))
        print(('find "xnat:out/xnat:file":', tree.find(
                'xnat:out/xnat:file', {'xnat': 'xnat'})))
        print(('find "xnat:out/xnat:file/nugget":', tree.find(
                'xnat:out/xnat:file/nugget', {'xnat': 'xnat'})))
        print(('findall "xnat:out/xnat:file":', tree.findall(
            'xnat:out/xnat:file', {'xnat': 'xnat'})))

    def test_xnat_get_cached_image_session(self):
        with XnatUtils.get_interface(host=host) as intf:

            cisess = XnatUtils.CachedImageSession(
                intf, proj_id, subj_id, sess_id)
            print(cisess)
            print((cisess.info()))
            for ciscan in cisess.scans():
                print((ciscan.info()))
                scanobj = ciscan.full_object()
                print(scanobj)
            for ciassr in cisess.assessors():
                print((ciassr.info()))
                for cirsrc in ciassr.out_resources():
                    print((cirsrc.info()))
                    asrinfo = ciassr.info()
                    rsrcobj = intf.select_assessor_resource(
                        asrinfo['project_id'], asrinfo['subject_id'],
                        asrinfo['session_id'], asrinfo['assessor_label'],
                        cirsrc.info()['label'])
                    print(rsrcobj)
            sessobj = cisess.full_object()
            print(sessobj)

    def test_assessor_out_resources(self):
        with XnatUtils.get_interface(host=host) as intf:
            cisess = XnatUtils.CachedImageSession(
                intf, proj_id, subj_id, sess_id)
            for asr in cisess.assessors():
                asrobj = intf.select_assessor(
                    proj_id, subj_id, sess_id, asr.label())
                print((asrobj.out_resources()))

    def test_xnat_get_full_object(self):
        with XnatUtils.get_interface(host=host) as intf:
            intf.connect()

            print(([x['name'] for x in intf.get_projects()]))

            print(([x['label'] for x in intf.get_subjects(proj_id)]))

            subj = intf.select(
                XnatUtils.InterfaceTemp.S_XPATH.format(
                    project=proj_id, subject=subj_id))
            print((subj.label()))
            print((subj.parent().label()))

    def test_xnat_parse_session(self):
        with XnatUtils.get_interface(host=host) as intf:
            intf.connect()

            SanityChecks.__prep_project(intf)

            yamldoc = YamlDoc().from_string(yamls.proc_a)
            csess = XnatUtils.CachedImageSession(
                intf, proj_id, subj_id, sess_id)

            ap = AutoProcessor(XnatUtils, yamldoc)
            ap.parse_session(ap)

    def test_xnat_has_inputs(self):
        with XnatUtils.get_interface(host=host) as intf:
            intf.connect()

            print((yamls.proc_a))
            SanityChecks.__prep_project(intf)

            yamldoc = YamlDoc().from_string(yamls.proc_a)
            ap = AutoProcessor(XnatUtils, yamldoc)
            csess = XnatUtils.CachedImageSession(
                intf, proj_id, subj_id, sess_id)

            results = []
            for cassr in csess.assessors():
                has, errors = ap.has_inputs(cassr)
                results.append((has, errors))

            print(results)

    def test_create_assessor_with_no_input(self):
        with XnatUtils.get_interface(host=host) as intf:
            intf.connect()

            e = intf.select_experiment(proj_id, subj_id, sess_id)
            if not e.exists():
                self.assertTrue(
                    False,
                    "Unexpected: {}//{}//{} does not exist".format(
                        proj_id, subj_id, sess_id
                    ))

            proc_a_params = {
                'xsitype': asrxsitype,
                'proctype': 'Proc_A_v1',
                'files': [
                    ('SEG', ['seg.gz'])
                ]
            }
            SessionTools.add_assessor(e,
                                      'proc1-x-subj1-x-sess1-x-2-Proc_A_v1',
                                      proc_a_params,
                                      'no_inputs')
